"""Test deck configuration provider."""

from typing import List, Set

import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.deck import load as load_deck
from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.types import (
    LocatingFeatures,
)
from opentrons_shared_data.module.types import ModuleOrientation

from opentrons.protocol_engine.errors import (
    AddressableAreaDoesNotExistError,
    CutoutDoesNotExistError,
    FixtureDoesNotExistError,
)
from opentrons.protocol_engine.resources import deck_configuration_provider as subject
from opentrons.protocol_engine.types import (
    AddressableArea,
    AddressableOffsetVector,
    AreaType,
    DeckPoint,
    Dimensions,
    PotentialCutoutFixture,
)
from opentrons.protocols.api_support.deck_type import (
    SHORT_TRASH_DECK,
    STANDARD_OT2_DECK,
    STANDARD_OT3_DECK,
)
from opentrons.types import DeckSlotName


@pytest.fixture(scope="session")
def ot2_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT2_DECK, 5)


@pytest.fixture(scope="session")
def ot2_short_trash_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(SHORT_TRASH_DECK, 5)


@pytest.fixture(scope="session")
def ot3_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT3_DECK, 5)


@pytest.mark.parametrize(
    ("cutout_id", "expected_deck_point", "deck_def"),
    [
        (
            "cutout5",
            DeckPoint(x=132.5, y=90.5, z=0.0),
            lazy_fixture("ot2_standard_deck_def"),
        ),
        (
            "cutout5",
            DeckPoint(x=132.5, y=90.5, z=0.0),
            lazy_fixture("ot2_short_trash_deck_def"),
        ),
        (
            "cutoutC2",
            DeckPoint(x=164.0, y=107, z=0.0),
            lazy_fixture("ot3_standard_deck_def"),
        ),
    ],
)
def test_get_cutout_position(
    cutout_id: str,
    expected_deck_point: DeckPoint,
    deck_def: DeckDefinitionV5,
) -> None:
    """It should get the deck position for the requested cutout id."""
    cutout_position = subject.get_cutout_position(cutout_id, deck_def)
    assert cutout_position == expected_deck_point


def test_get_cutout_position_raises(
    ot3_standard_deck_def: DeckDefinitionV5,
) -> None:
    """It should raise if there is no cutout with that ID in the deck definition."""
    with pytest.raises(CutoutDoesNotExistError):
        subject.get_cutout_position("theFunCutout", ot3_standard_deck_def)


@pytest.mark.parametrize(
    ("cutout_fixture_id", "expected_display_name", "deck_def"),
    [
        ("singleStandardSlot", "Standard Slot", lazy_fixture("ot2_standard_deck_def")),
        (
            "singleStandardSlot",
            "Standard Slot",
            lazy_fixture("ot2_short_trash_deck_def"),
        ),
        (
            "singleRightSlot",
            "Standard Slot Right",
            lazy_fixture("ot3_standard_deck_def"),
        ),
    ],
)
def test_get_cutout_fixture(
    cutout_fixture_id: str,
    expected_display_name: str,
    deck_def: DeckDefinitionV5,
) -> None:
    """It should get the cutout fixture given the cutout fixture id."""
    cutout_fixture = subject.get_cutout_fixture(cutout_fixture_id, deck_def)
    assert cutout_fixture["displayName"] == expected_display_name


def test_get_cutout_fixture_raises(
    ot3_standard_deck_def: DeckDefinitionV5,
) -> None:
    """It should raise if the given cutout fixture id does not exist."""
    with pytest.raises(FixtureDoesNotExistError):
        subject.get_cutout_fixture("theFunFixture", ot3_standard_deck_def)


@pytest.mark.parametrize(
    ("cutout_fixture_id", "cutout_id", "expected_areas", "deck_def"),
    [
        (
            "singleStandardSlot",
            "cutout1",
            ["1"],
            lazy_fixture("ot2_standard_deck_def"),
        ),
        (
            "singleStandardSlot",
            "cutout1",
            ["1"],
            lazy_fixture("ot2_short_trash_deck_def"),
        ),
        (
            "stagingAreaRightSlot",
            "cutoutD3",
            ["D3", "D4"],
            lazy_fixture("ot3_standard_deck_def"),
        ),
    ],
)
def test_get_provided_addressable_area_names(
    cutout_fixture_id: str,
    cutout_id: str,
    expected_areas: List[str],
    deck_def: DeckDefinitionV5,
) -> None:
    """It should get the provided addressable area for the cutout fixture and cutout."""
    provided_addressable_areas = subject.get_provided_addressable_area_names(
        cutout_fixture_id, cutout_id, deck_def
    )
    assert provided_addressable_areas == expected_areas


@pytest.mark.parametrize(
    (
        "addressable_area_name",
        "expected_cutout_id",
        "expected_potential_fixtures",
        "deck_def",
    ),
    [
        (
            "3",
            "cutout3",
            {
                PotentialCutoutFixture(
                    cutout_id="cutout3",
                    cutout_fixture_id="singleStandardSlot",
                    provided_addressable_areas=frozenset({"3"}),
                )
            },
            lazy_fixture("ot2_standard_deck_def"),
        ),
        (
            "3",
            "cutout3",
            {
                PotentialCutoutFixture(
                    cutout_id="cutout3",
                    cutout_fixture_id="singleStandardSlot",
                    provided_addressable_areas=frozenset({"3"}),
                )
            },
            lazy_fixture("ot2_short_trash_deck_def"),
        ),
        (
            "D3",
            "cutoutD3",
            {
                PotentialCutoutFixture(
                    cutout_id="cutoutD3",
                    cutout_fixture_id="singleRightSlot",
                    provided_addressable_areas=frozenset({"D3"}),
                ),
                PotentialCutoutFixture(
                    cutout_id="cutoutD3",
                    cutout_fixture_id="stagingAreaRightSlot",
                    provided_addressable_areas=frozenset({"D3", "D4"}),
                ),
                PotentialCutoutFixture(
                    cutout_id="cutoutD3",
                    cutout_fixture_id="flexStackerModuleV1",
                    provided_addressable_areas=frozenset(
                        {"D3", "flexStackerModuleV1D4"}
                    ),
                ),
            },
            lazy_fixture("ot3_standard_deck_def"),
        ),
        (
            "flexStackerModuleV1D4",
            "cutoutD3",
            {
                PotentialCutoutFixture(
                    cutout_id="cutoutD3",
                    cutout_fixture_id="flexStackerModuleV1",
                    provided_addressable_areas=frozenset(
                        {"D3", "flexStackerModuleV1D4"}
                    ),
                ),
                PotentialCutoutFixture(
                    cutout_id="cutoutD3",
                    cutout_fixture_id="flexStackerModuleV1WithMagneticBlockV1",
                    provided_addressable_areas=frozenset(
                        {"flexStackerModuleV1D4", "magneticBlockV1D3"}
                    ),
                ),
                PotentialCutoutFixture(
                    cutout_id="cutoutD3",
                    cutout_fixture_id="flexStackerModuleV1WithWasteChuteRightAdapterCovered",
                    provided_addressable_areas=frozenset(
                        {
                            "1ChannelWasteChute",
                            "8ChannelWasteChute",
                            "flexStackerModuleV1D4",
                        }
                    ),
                ),
                PotentialCutoutFixture(
                    cutout_id="cutoutD3",
                    cutout_fixture_id="flexStackerModuleV1WithWasteChuteRightAdapterNoCover",
                    provided_addressable_areas=frozenset(
                        {
                            "1ChannelWasteChute",
                            "8ChannelWasteChute",
                            "96ChannelWasteChute",
                            "gripperWasteChute",
                            "flexStackerModuleV1D4",
                        }
                    ),
                ),
            },
            lazy_fixture("ot3_standard_deck_def"),
        ),
    ],
)
def test_get_potential_cutout_fixtures(
    addressable_area_name: str,
    expected_cutout_id: str,
    expected_potential_fixtures: Set[PotentialCutoutFixture],
    deck_def: DeckDefinitionV5,
) -> None:
    """It should get a cutout id and a set of potential cutout fixtures for an addressable area name."""
    cutout_id, potential_fixtures = subject.get_potential_cutout_fixtures(
        addressable_area_name, deck_def
    )
    assert cutout_id == expected_cutout_id
    assert potential_fixtures == expected_potential_fixtures


def test_get_potential_cutout_fixtures_raises(
    ot3_standard_deck_def: DeckDefinitionV5,
) -> None:
    """It should raise if there is no fixtures that provide the requested area."""
    with pytest.raises(AddressableAreaDoesNotExistError):
        subject.get_potential_cutout_fixtures("theFunArea", ot3_standard_deck_def)


# TODO put in fixed trash for OT2 decks
@pytest.mark.parametrize(
    ("addressable_area_name", "expected_addressable_area", "deck_def"),
    [
        (
            "1",
            AddressableArea(
                area_name="1",
                area_type=AreaType.SLOT,
                base_slot=DeckSlotName.SLOT_1,
                display_name="Slot 1",
                bounding_box=Dimensions(x=128.0, y=86.0, z=0),
                position=AddressableOffsetVector(x=1, y=2, z=3),
                features=LocatingFeatures(
                    springDirectionalForceAsParent="backLeftBottom"
                ),
                compatible_module_types=[
                    "magneticModuleType",
                    "temperatureModuleType",
                    "heaterShakerModuleType",
                ],
                mating_surface_unit_vector=[-1, 1, -1],
                orientation=ModuleOrientation.NOT_APPLICABLE,
            ),
            lazy_fixture("ot2_standard_deck_def"),
        ),
        (
            "1",
            AddressableArea(
                area_name="1",
                area_type=AreaType.SLOT,
                base_slot=DeckSlotName.SLOT_1,
                display_name="Slot 1",
                bounding_box=Dimensions(x=128.0, y=86.0, z=0),
                position=AddressableOffsetVector(x=1, y=2, z=3),
                compatible_module_types=[
                    "magneticModuleType",
                    "temperatureModuleType",
                    "heaterShakerModuleType",
                ],
                features=LocatingFeatures(
                    springDirectionalForceAsParent="backLeftBottom"
                ),
                mating_surface_unit_vector=[-1, 1, -1],
                orientation=ModuleOrientation.NOT_APPLICABLE,
            ),
            lazy_fixture("ot2_short_trash_deck_def"),
        ),
        (
            "D1",
            AddressableArea(
                area_name="D1",
                area_type=AreaType.SLOT,
                base_slot=DeckSlotName.SLOT_D1,
                display_name="Slot D1",
                bounding_box=Dimensions(x=128.0, y=86.0, z=0),
                position=AddressableOffsetVector(x=1, y=2, z=3),
                compatible_module_types=[],
                features=LocatingFeatures(
                    springDirectionalForceAsParent="backLeftBottom"
                ),
                mating_surface_unit_vector=[-1, 1, -1],
                orientation=ModuleOrientation.NOT_APPLICABLE,
            ),
            lazy_fixture("ot3_standard_deck_def"),
        ),
        (
            "movableTrashB3",
            AddressableArea(
                area_name="movableTrashB3",
                area_type=AreaType.MOVABLE_TRASH,
                base_slot=DeckSlotName.SLOT_B3,
                display_name="Trash Bin in B3",
                bounding_box=Dimensions(x=225, y=78, z=40),
                position=AddressableOffsetVector(x=-5.25, y=6, z=3),
                compatible_module_types=[],
                features=LocatingFeatures(),
                mating_surface_unit_vector=None,
                orientation=ModuleOrientation.NOT_APPLICABLE,
            ),
            lazy_fixture("ot3_standard_deck_def"),
        ),
        (
            "gripperWasteChute",
            AddressableArea(
                area_name="gripperWasteChute",
                area_type=AreaType.WASTE_CHUTE,
                base_slot=DeckSlotName.SLOT_D3,
                display_name="Waste Chute",
                bounding_box=Dimensions(x=0, y=0, z=0),
                position=AddressableOffsetVector(x=65, y=31, z=139.5),
                compatible_module_types=[],
                features=LocatingFeatures(),
                mating_surface_unit_vector=None,
                orientation=ModuleOrientation.NOT_APPLICABLE,
            ),
            lazy_fixture("ot3_standard_deck_def"),
        ),
    ],
)
def test_get_addressable_area_from_name(
    addressable_area_name: str,
    expected_addressable_area: AddressableArea,
    deck_def: DeckDefinitionV5,
) -> None:
    """It should get the deck position for the requested cutout id."""
    addressable_area = subject.get_addressable_area_from_name(
        addressable_area_name, DeckPoint(x=1, y=2, z=3), deck_def
    )
    assert addressable_area == expected_addressable_area


def test_get_addressable_area_from_name_raises(
    ot3_standard_deck_def: DeckDefinitionV5,
) -> None:
    """It should raise if there is no addressable area by that name in the deck."""
    with pytest.raises(AddressableAreaDoesNotExistError):
        subject.get_addressable_area_from_name(
            "theFunArea",
            DeckPoint(x=1, y=2, z=3),
            ot3_standard_deck_def,
        )



"""Task state tracking."""

from dataclasses import dataclass
from itertools import chain
from typing import Iterable

from ..actions import (
    Action,
    FinishTaskAction,
    StartTaskAction,
    get_state_updates,
)
from ..types import FinishedTask, Task, TaskSummary
from ._abstract_store import HandlesActions, HasState
from opentrons.protocol_engine.errors.exceptions import NoTaskFoundError
from opentrons.protocol_engine.state import update_types


@dataclass
class TaskState:
    """Task state tracking."""

    current_tasks_by_id: dict[str, Task]
    finished_tasks_by_id: dict[str, FinishedTask]


class TaskStore(HasState[TaskState], HandlesActions):
    """Stores tasks."""

    _state: TaskState

    def __init__(self) -> None:
        """Initialize a TaskStore."""
        self._state = TaskState(current_tasks_by_id={}, finished_tasks_by_id={})

    def _handle_state_update(self, state_update: update_types.StateUpdate) -> None:
        """Handle a state update."""
        return

    def _handle_start_task_action(self, action: StartTaskAction) -> None:
        self._state.current_tasks_by_id[action.task.id] = action.task

    def _handle_finish_task_action(self, action: FinishTaskAction) -> None:
        task = self._state.current_tasks_by_id[action.task_id]
        self._state.finished_tasks_by_id[action.task_id] = FinishedTask(
            id=task.id,
            createdAt=task.createdAt,
            finishedAt=action.finished_at,
            error=action.error,
        )
        del self._state.current_tasks_by_id[action.task_id]

    def handle_action(self, action: Action) -> None:
        """Modify the state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._handle_state_update(state_update)
        match action:
            case StartTaskAction():
                self._handle_start_task_action(action)
            case FinishTaskAction():
                self._handle_finish_task_action(action)
            case _:
                pass


class TaskView:
    """Read-only task state view."""

    _state: TaskState

    def __init__(self, state: TaskState) -> None:
        """Initialize a TaskView."""
        self._state = state

    def get_current(self, id: str) -> Task:
        """Get a task by ID."""
        try:
            return self._state.current_tasks_by_id[id]
        except KeyError as e:
            raise NoTaskFoundError(f"No current task with ID {id}") from e

    def get_all_current(self) -> list[Task]:
        """Get all currently running tasks."""
        return [task for task in self._state.current_tasks_by_id.values()]

    def get_finished(self, id: str) -> FinishedTask:
        """Get a finished task by ID."""
        try:
            return self._state.finished_tasks_by_id[id]
        except KeyError as e:
            raise NoTaskFoundError(f"No finished task with ID {id}") from e

    def get(self, id: str) -> Task | FinishedTask:
        """Get a single task by id."""
        if id in self._state.current_tasks_by_id:
            return self._state.current_tasks_by_id[id]
        elif id in self._state.finished_tasks_by_id:
            return self._state.finished_tasks_by_id[id]
        else:
            raise NoTaskFoundError(message=f"Task {id} not found.")

    def get_summary(self) -> list[TaskSummary]:
        """Get a summary of all tasks."""
        return [
            TaskSummary(
                id=task_id,
                createdAt=task.createdAt,
                finishedAt=getattr(task, "finishedAt", None),
                error=getattr(task, "error", None),
            )
            for task_id, task in chain(
                self._state.current_tasks_by_id.items(),
                self._state.finished_tasks_by_id.items(),
            )
        ]

    def all_tasks_finished_or_any_task_failed(self, task_ids: Iterable[str]) -> bool:
        """Implements wait semantics of asyncio.gather(return_exceptions = False).

        This returns true when any of the following are true:
        - All tasks in task_ids are complete with or without an error
        - Any task in task_ids is complete with an error.

        NOTE: Does not raise the error that the errored task has.
        """
        finished = set(self._state.finished_tasks_by_id.keys())
        task_ids = set(task_ids)
        if task_ids.issubset(finished):
            return True
        if self.get_failed_tasks(task_ids):
            return True
        return False

    def get_failed_tasks(self, task_ids: Iterable[str]) -> list[str]:
        """Return a list of failed task ids of the ones that were passed."""
        failed_tasks: list[str] = []
        for task_id in task_ids:
            task = self._state.finished_tasks_by_id.get(task_id, None)
            if task and task.error:
                failed_tasks.append(task_id)
        return failed_tasks



"""Tests for tip state store and selectors."""

from typing import Optional

import pytest

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    LabwareDefinition2,
)
from opentrons_shared_data.labware.labware_definition import (
    Parameters2 as LabwareDefinition2Parameters,
)
from opentrons_shared_data.pipette.pipette_definition import (
    AvailableSensorDefinition,
    ValidNozzleMaps,
)
from opentrons_shared_data.pipette.types import PipetteNameType

from ..pipette_fixtures import (
    NINETY_SIX_COLS,
    NINETY_SIX_MAP,
    NINETY_SIX_ROWS,
    get_default_nozzle_map,
)
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine import actions, commands
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.tips import TipStore, TipView
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    DeckSlotLocation,
    TipRackWellState,
)
from opentrons.types import DeckSlotName

_tip_rack_parameters = LabwareDefinition2Parameters.model_construct(isTiprack=True)  # type: ignore[call-arg]


@pytest.fixture
def available_sensors() -> AvailableSensorDefinition:
    """Provide a list of sensors."""
    return AvailableSensorDefinition(sensors=["pressure", "capacitive", "environment"])


@pytest.fixture
def subject() -> TipStore:
    """Get a TipStore test subject."""
    return TipStore()


@pytest.fixture
def labware_definition() -> LabwareDefinition:
    """Get a labware definition value object."""
    return LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        ordering=[
            ["A1", "B1", "C1", "D1", "E1", "F1", "G1", "H1"],
            ["A2", "B2", "C2", "D2", "E2", "F2", "G2", "H2"],
            ["A3", "B3", "C3", "D3", "E3", "F3", "G3", "H3"],
            ["A4", "B4", "C4", "D4", "E4", "F4", "G4", "H4"],
            ["A5", "B5", "C5", "D5", "E5", "F5", "G5", "H5"],
            ["A6", "B6", "C6", "D6", "E6", "F6", "G6", "H6"],
            ["A7", "B7", "C7", "D7", "E7", "F7", "G7", "H7"],
            ["A8", "B8", "C8", "D8", "E8", "F8", "G8", "H8"],
            ["A9", "B9", "C9", "D9", "E9", "F9", "G9", "H9"],
            ["A10", "B10", "C10", "D10", "E10", "F10", "G10", "H10"],
            ["A11", "B11", "C11", "D11", "E11", "F11", "G11", "H11"],
            ["A12", "B12", "C12", "D12", "E12", "F12", "G12", "H12"],
        ],
        parameters=_tip_rack_parameters,
    )


@pytest.fixture
def load_labware_action(
    labware_definition: LabwareDefinition,
) -> actions.SucceedCommandAction:
    """Get a load labware command value object."""
    return actions.SucceedCommandAction(
        command=_dummy_command(),
        state_update=update_types.StateUpdate(
            loaded_labware=update_types.LoadedLabwareUpdate(
                labware_id="cool-labware",
                definition=labware_definition,
                new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
                display_name=None,
                offset_id=None,
            )
        ),
    )


def _dummy_command() -> commands.Command:
    """Return a placeholder command."""
    return commands.Comment.model_construct()  # type: ignore[call-arg]


@pytest.mark.parametrize(
    "labware_definition",
    [LabwareDefinition2.model_construct(ordering=[], parameters=_tip_rack_parameters)],  # type: ignore[call-arg]
)
def test_get_next_tip_returns_none(
    load_labware_action: actions.SucceedCommandAction,
    subject: TipStore,
) -> None:
    """It should start at the first tip in the labware."""
    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=1,
        starting_tip_name=None,
        nozzle_map=None,
    )

    assert result is None


@pytest.mark.parametrize("input_tip_amount", [1, 8, 96])
def test_get_next_tip_returns_first_tip(
    load_labware_action: actions.SucceedCommandAction,
    subject: TipStore,
    input_tip_amount: int,
) -> None:
    """It should start at the first tip in the labware."""
    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=input_tip_amount,
        starting_tip_name=None,
        nozzle_map=None,
    )

    assert result == "A1"


@pytest.mark.parametrize("input_tip_amount, result_well_name", [(1, "B1"), (8, "A2")])
def test_get_next_tip_used_starting_tip(
    load_labware_action: actions.SucceedCommandAction,
    subject: TipStore,
    input_tip_amount: int,
    result_well_name: str,
) -> None:
    """It should start searching at the given starting tip."""
    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=input_tip_amount,
        starting_tip_name="B1",
        nozzle_map=None,
    )

    assert result == result_well_name


@pytest.mark.parametrize(
    "input_tip_amount, get_next_tip_tips, input_starting_tip, result_well_name",
    [
        (1, 8, "A2", "A2"),
        (1, 1, "A2", "A2"),
        (8, 8, "B2", "A3"),
        (1, 8, "A1", "A2"),
        (8, 1, "D1", "A2"),
        (1, 96, "A1", None),
        (1, 8, None, "A2"),
        (8, 1, "D1", "A2"),
        (1, 96, None, None),
    ],
)
def test_get_next_tip_skips_picked_up_tip(
    load_labware_action: actions.SucceedCommandAction,
    subject: TipStore,
    input_tip_amount: int,
    get_next_tip_tips: int,
    input_starting_tip: Optional[str],
    result_well_name: Optional[str],
) -> None:
    """It should get the next tip in the column if one has been picked up."""
    subject.handle_action(load_labware_action)

    if input_starting_tip is not None:
        if input_tip_amount == 1:
            pipette_name_type = PipetteNameType.P300_SINGLE_GEN2
        elif input_tip_amount == 8:
            pipette_name_type = PipetteNameType.P300_MULTI_GEN2
        else:
            pipette_name_type = PipetteNameType.P1000_96
    else:
        if get_next_tip_tips == 1:
            pipette_name_type = PipetteNameType.P300_SINGLE_GEN2
        elif get_next_tip_tips == 8:
            pipette_name_type = PipetteNameType.P300_MULTI_GEN2
        else:
            pipette_name_type = PipetteNameType.P1000_96

    nozzle_map = get_default_nozzle_map(pipette_name_type)

    pick_up_tip_state_update = update_types.StateUpdate(
        tips_state=update_types.TipsStateUpdate(
            tip_state=TipRackWellState.EMPTY,
            labware_id="cool-labware",
            well_names=TipView(subject.state).compute_tips_to_mark_as_used_or_empty(
                labware_id="cool-labware", well_name="A1", nozzle_map=nozzle_map
            ),
        )
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            command=_dummy_command(),
            state_update=pick_up_tip_state_update,
        )
    )

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=get_next_tip_tips,
        starting_tip_name=input_starting_tip,
        nozzle_map=nozzle_map,
    )

    assert result == result_well_name


def test_get_next_tip_with_starting_tip(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """It should return the starting tip, and then the following tip after that."""
    nozzle_map = get_default_nozzle_map(PipetteNameType.P300_SINGLE_GEN2)

    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=1,
        starting_tip_name="B2",
        nozzle_map=nozzle_map,
    )
    assert result == "B2"

    pick_up_tip_state_update = update_types.StateUpdate(
        tips_state=update_types.TipsStateUpdate(
            tip_state=TipRackWellState.EMPTY,
            labware_id="cool-labware",
            well_names=TipView(subject.state).compute_tips_to_mark_as_used_or_empty(
                labware_id="cool-labware", well_name="B2", nozzle_map=nozzle_map
            ),
        )
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            command=_dummy_command(),
            state_update=pick_up_tip_state_update,
        )
    )

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=1,
        starting_tip_name="B2",
        nozzle_map=nozzle_map,
    )
    assert result == "C2"


def test_get_next_tip_with_starting_tip_8_channel(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """It should return the starting tip, and then the following tip after that."""
    nozzle_map = get_default_nozzle_map(PipetteNameType.P300_MULTI_GEN2)

    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=8,
        starting_tip_name="A2",
        nozzle_map=None,
    )
    assert result == "A2"

    pick_up_tip_state_update = update_types.StateUpdate(
        tips_state=update_types.TipsStateUpdate(
            tip_state=TipRackWellState.EMPTY,
            labware_id="cool-labware",
            well_names=TipView(subject.state).compute_tips_to_mark_as_used_or_empty(
                labware_id="cool-labware", well_name="A2", nozzle_map=nozzle_map
            ),
        )
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            command=_dummy_command(),
            state_update=pick_up_tip_state_update,
        )
    )

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=8,
        starting_tip_name="A2",
        nozzle_map=None,
    )

    assert result == "A3"


def test_get_next_tip_with_1_channel_followed_by_8_channel(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """It should return the first tip of column 2 for the 8 channel after performing a single tip pickup on column 1."""
    nozzle_map_1_channel = get_default_nozzle_map(PipetteNameType.P300_SINGLE_GEN2)
    nozzle_map_8_channel = get_default_nozzle_map(PipetteNameType.P300_MULTI_GEN2)

    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=1,
        starting_tip_name=None,
        nozzle_map=nozzle_map_1_channel,
    )
    assert result == "A1"

    pick_up_tip_1_channel_state_update = update_types.StateUpdate(
        tips_state=update_types.TipsStateUpdate(
            tip_state=TipRackWellState.EMPTY,
            labware_id="cool-labware",
            well_names=TipView(subject.state).compute_tips_to_mark_as_used_or_empty(
                labware_id="cool-labware",
                well_name="A1",
                nozzle_map=nozzle_map_1_channel,
            ),
        )
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            command=_dummy_command(),
            state_update=pick_up_tip_1_channel_state_update,
        )
    )

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=8,
        starting_tip_name=None,
        nozzle_map=nozzle_map_8_channel,
    )
    assert result == "A2"


def test_get_next_tip_with_starting_tip_out_of_tips(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """It should return the starting tip of H12 and then None after that."""
    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=1,
        starting_tip_name="H12",
        nozzle_map=None,
    )
    assert result == "H12"

    pick_up_tip_state_update = update_types.StateUpdate(
        tips_state=update_types.TipsStateUpdate(
            tip_state=TipRackWellState.EMPTY,
            labware_id="cool-labware",
            well_names=TipView(subject.state).compute_tips_to_mark_as_used_or_empty(
                labware_id="cool-labware",
                well_name="H12",
                nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE_GEN2),
            ),
        )
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            command=_dummy_command(),
            state_update=pick_up_tip_state_update,
        )
    )

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=1,
        starting_tip_name="H12",
        nozzle_map=None,
    )

    assert result is None


def test_get_next_tip_with_column_and_starting_tip(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """It should return the first tip in a column, taking starting tip into account."""
    subject.handle_action(load_labware_action)

    result = TipView(subject.state).get_next_tip(
        labware_id="cool-labware",
        num_tips=8,
        starting_tip_name="D1",
        nozzle_map=None,
    )

    assert result == "A2"


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[["A1"]],
            parameters=LabwareDefinition2Parameters.model_construct(isTiprack=False),  # type: ignore[call-arg]
        )
    ],
)
def test_has_tip_not_tip_rack(
    load_labware_action: actions.SucceedCommandAction, subject: TipStore
) -> None:
    """It should return False if labware isn't a tip rack."""
    subject.handle_action(load_labware_action)

    result = TipView(state=subject.state).has_clean_tip("cool-labware", "A1")

    assert result is False


def test_has_tip_tip_rack(
    load_labware_action: actions.SucceedCommandAction, subject: TipStore
) -> None:
    """It should return False if labware isn't a tip rack."""
    subject.handle_action(load_labware_action)

    result = TipView(state=subject.state).has_clean_tip("cool-labware", "A1")

    assert result is True


def test_next_tip_automatic_tip_tracking_with_partial_configurations(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """Test tip tracking logic using multiple pipette configurations."""
    subject.handle_action(load_labware_action)

    def _assert_and_pickup(expected_next_tip: str, nozzle_map: NozzleMap) -> None:
        result = TipView(subject.state).get_next_tip(
            labware_id="cool-labware",
            num_tips=0,
            starting_tip_name=None,
            nozzle_map=nozzle_map,
        )
        assert result is not None and result == expected_next_tip

        pick_up_tip_state_update = update_types.StateUpdate(
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.EMPTY,
                labware_id="cool-labware",
                well_names=TipView(subject.state).compute_tips_to_mark_as_used_or_empty(
                    labware_id="cool-labware", well_name=result, nozzle_map=nozzle_map
                ),
            )
        )

        subject.handle_action(
            actions.SucceedCommandAction(
                command=_dummy_command(),
                state_update=pick_up_tip_state_update,
            )
        )

    def _build_nozzle_map(start: str, back_l: str, front_r: str) -> NozzleMap:
        return NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle=start,
            back_left_nozzle=back_l,
            front_right_nozzle=front_r,
            valid_nozzle_maps=ValidNozzleMaps(
                maps={
                    "A1": ["A1"],
                    "H1": ["H1"],
                    "A12": ["A12"],
                    "H12": ["H12"],
                    "A1_H3": [
                        "A1",
                        "A2",
                        "A3",
                        "B1",
                        "B2",
                        "B3",
                        "C1",
                        "C2",
                        "C3",
                        "D1",
                        "D2",
                        "D3",
                        "E1",
                        "E2",
                        "E3",
                        "F1",
                        "F2",
                        "F3",
                        "G1",
                        "G2",
                        "G3",
                        "H1",
                        "H2",
                        "H3",
                    ],
                    "A1_F2": [
                        "A1",
                        "A2",
                        "B1",
                        "B2",
                        "C1",
                        "C2",
                        "D1",
                        "D2",
                        "E1",
                        "E2",
                        "F1",
                        "F2",
                    ],
                }
            ),
        )

    pipette_nozzle_map = _build_nozzle_map("A1", "A1", "H3")
    _assert_and_pickup("A10", pipette_nozzle_map)
    pipette_nozzle_map = _build_nozzle_map("A1", "A1", "F2")
    _assert_and_pickup("C8", pipette_nozzle_map)

    # Configure to single tip pickups
    pipette_nozzle_map = _build_nozzle_map("H12", "H12", "H12")
    _assert_and_pickup("A1", pipette_nozzle_map)
    pipette_nozzle_map = _build_nozzle_map("H1", "H1", "H1")
    _assert_and_pickup("A9", pipette_nozzle_map)
    pipette_nozzle_map = _build_nozzle_map("A12", "A12", "A12")
    _assert_and_pickup("H1", pipette_nozzle_map)
    pipette_nozzle_map = _build_nozzle_map("A1", "A1", "A1")
    _assert_and_pickup("B9", pipette_nozzle_map)


def test_next_tip_automatic_tip_tracking_tiprack_limits(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """Ensure once a tip rack is consumed it returns None when consuming tips using multiple pipette configurations."""
    # Load labware
    subject.handle_action(load_labware_action)

    def _get_next_and_pickup(nozzle_map: NozzleMap) -> str | None:
        result = TipView(subject.state).get_next_tip(
            labware_id="cool-labware",
            num_tips=0,
            starting_tip_name=None,
            nozzle_map=nozzle_map,
        )
        if result is not None:
            pick_up_tip_state_update = update_types.StateUpdate(
                tips_state=update_types.TipsStateUpdate(
                    tip_state=TipRackWellState.EMPTY,
                    labware_id="cool-labware",
                    well_names=TipView(
                        subject.state
                    ).compute_tips_to_mark_as_used_or_empty(
                        labware_id="cool-labware",
                        well_name=result,
                        nozzle_map=nozzle_map,
                    ),
                )
            )

            subject.handle_action(
                actions.SucceedCommandAction(
                    command=_dummy_command(),
                    state_update=pick_up_tip_state_update,
                )
            )

        return result

    def _build_nozzle_map(start: str, back_l: str, front_r: str) -> NozzleMap:
        return NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle=start,
            back_left_nozzle=back_l,
            front_right_nozzle=front_r,
            valid_nozzle_maps=ValidNozzleMaps(
                maps={
                    "A1": ["A1"],
                    "H1": ["H1"],
                    "A12": ["A12"],
                    "H12": ["H12"],
                    "Full": sum(
                        [
                            NINETY_SIX_ROWS["A"],
                            NINETY_SIX_ROWS["B"],
                            NINETY_SIX_ROWS["C"],
                            NINETY_SIX_ROWS["D"],
                            NINETY_SIX_ROWS["E"],
                            NINETY_SIX_ROWS["F"],
                            NINETY_SIX_ROWS["G"],
                            NINETY_SIX_ROWS["H"],
                        ],
                        [],
                    ),
                }
            ),
        )

    pipette_nozzle_map = _build_nozzle_map("A1", "A1", "A1")
    for _ in range(96):
        assert _get_next_and_pickup(pipette_nozzle_map) is not None
    assert _get_next_and_pickup(pipette_nozzle_map) is None

    subject.handle_action(load_labware_action)
    pipette_nozzle_map = _build_nozzle_map("A12", "A12", "A12")
    for _ in range(96):
        assert _get_next_and_pickup(pipette_nozzle_map) is not None
    assert _get_next_and_pickup(pipette_nozzle_map) is None

    subject.handle_action(load_labware_action)
    pipette_nozzle_map = _build_nozzle_map("H1", "H1", "H1")
    for _ in range(96):
        assert _get_next_and_pickup(pipette_nozzle_map) is not None
    assert _get_next_and_pickup(pipette_nozzle_map) is None

    subject.handle_action(load_labware_action)
    pipette_nozzle_map = _build_nozzle_map("H12", "H12", "H12")
    for _ in range(96):
        assert _get_next_and_pickup(pipette_nozzle_map) is not None
    assert _get_next_and_pickup(pipette_nozzle_map) is None


def test_96_column_after_row_returns_none(
    subject: TipStore,
    load_labware_action: actions.SucceedCommandAction,
) -> None:
    """It should return None when there's no valid columns to pick up."""
    subject.handle_action(load_labware_action)

    def _get_next_and_pickup(nozzle_map: NozzleMap) -> str | None:
        result = TipView(subject.state).get_next_tip(
            labware_id="cool-labware",
            num_tips=0,
            starting_tip_name=None,
            nozzle_map=nozzle_map,
        )
        if result is not None:
            pick_up_tip_state_update = update_types.StateUpdate(
                tips_state=update_types.TipsStateUpdate(
                    tip_state=TipRackWellState.EMPTY,
                    labware_id="cool-labware",
                    well_names=TipView(
                        subject.state
                    ).compute_tips_to_mark_as_used_or_empty(
                        labware_id="cool-labware",
                        well_name=result,
                        nozzle_map=nozzle_map,
                    ),
                )
            )

            subject.handle_action(
                actions.SucceedCommandAction(
                    command=_dummy_command(),
                    state_update=pick_up_tip_state_update,
                )
            )

        return result

    row_nozzle_map = NozzleMap.build(
        physical_nozzles=NINETY_SIX_MAP,
        physical_rows=NINETY_SIX_ROWS,
        physical_columns=NINETY_SIX_COLS,
        starting_nozzle="A1",
        back_left_nozzle="A1",
        front_right_nozzle="A12",
        valid_nozzle_maps=ValidNozzleMaps(maps={"RowA": NINETY_SIX_ROWS["A"]}),
    )
    assert _get_next_and_pickup(row_nozzle_map) is not None

    col_nozzle_map = NozzleMap.build(
        physical_nozzles=NINETY_SIX_MAP,
        physical_rows=NINETY_SIX_ROWS,
        physical_columns=NINETY_SIX_COLS,
        starting_nozzle="A1",
        back_left_nozzle="A1",
        front_right_nozzle="H1",
        valid_nozzle_maps=ValidNozzleMaps(maps={"Column1": NINETY_SIX_COLS["1"]}),
    )
    assert _get_next_and_pickup(col_nozzle_map) is None


def test_handle_batch_labware_loaded_update(
    subject: TipStore, labware_definition: LabwareDefinition
) -> None:
    """It should consume batch_loaded_labware updates."""
    subject.handle_action(
        actions.SucceedCommandAction(
            command=_dummy_command(),
            state_update=update_types.StateUpdate(
                batch_loaded_labware=update_types.BatchLoadedLabwareUpdate(
                    new_locations_by_id={
                        "some-labware-id": DeckSlotLocation(
                            slotName=DeckSlotName.SLOT_1
                        ),
                        "some-other-labware-id": OFF_DECK_LOCATION,
                    },
                    offset_ids_by_id={
                        "some-labware-id": "some-offset-id",
                        "some-other-labware-id": None,
                    },
                    display_names_by_id={
                        "some-labware-id": "some-display-name",
                        "some-other-labware-id": None,
                    },
                    definitions_by_id={
                        "some-labware-id": labware_definition,
                        "some-other-labware-id": labware_definition,
                    },
                )
            ),
        )
    )

    # The use of has_clean_tip() is arbitrary here. We just need anything that can make
    # sure each labware in the batch has actually been ingested.
    assert TipView(subject.state).has_clean_tip("some-labware-id", "A1")
    assert TipView(subject.state).has_clean_tip("some-other-labware-id", "A1")



from abc import ABC, abstractmethod
from datetime import datetime
from typing import TypeVar


class AbstractTaskCore(ABC):
    @abstractmethod
    def get_created_at_timestamp(self) -> datetime:
        """Get the createdAt timestamp of the task."""
        ...

    @abstractmethod
    def is_done(self) -> bool:
        """Returns ``True`` if the task is done."""
        ...

    @abstractmethod
    def is_started(self) -> bool:
        """Returns ``True`` if the task has started."""
        ...

    @abstractmethod
    def get_finished_at_timestamp(self) -> datetime | None:
        """The timestamp of the when the task finished.

        Returns ``None`` if the task hasn't finished yet.
        """
        ...


TaskCoreType = TypeVar("TaskCoreType", bound=AbstractTaskCore)



from typing import Dict

import pytest

from opentrons.drivers import utils
from opentrons.drivers.types import (
    RPM,
    HeaterShakerLabwareLatchStatus,
    PlateTemperature,
    Temperature,
)


@pytest.mark.parametrize(
    argnames=["input_str", "expected_result"],
    argvalues=[
        [
            "version:123-2 serial:serial_v model:m",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
        [
            "serial:serial_v model:m version:123-2 ",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
        [
            "   serial:serial_v model:m    version:123-2   ",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
        [
            "something:extra serial:serial_v model:m version:123-2",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
    ],
)
def test_parse_device_information_success(
    input_str: str, expected_result: Dict[str, str]
) -> None:
    """Test parse device information."""
    assert utils.parse_device_information(input_str) == expected_result


@pytest.mark.parametrize(
    argnames=["input_str"],
    argvalues=[
        ["version:123 serial:serial_v"],
        ["version:123 serialg:serial_v model:123"],
        [""],
    ],
)
def test_parse_device_information_failure(input_str: str) -> None:
    """Test parse device information."""
    with pytest.raises(utils.ParseError):
        utils.parse_device_information(input_str)


@pytest.mark.parametrize(
    argnames=["input_str", "expected_result"],
    argvalues=[
        [
            "FW:123-2 SerialNo:serial_v HW:m",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
        [
            "SerialNo:serial_v HW:m FW:123-2 ",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
        [
            "   SerialNo:serial_v HW:m    FW:123-2   ",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
        [
            "something:extra SerialNo:serial_v HW:m FW:123-2",
            {"version": "123-2", "serial": "serial_v", "model": "m"},
        ],
    ],
)
def test_parse_hs_device_information_success(
    input_str: str, expected_result: Dict[str, str]
) -> None:
    """Test parse device information."""
    assert utils.parse_hs_device_information(input_str) == expected_result


@pytest.mark.parametrize(
    argnames=["input_str"],
    argvalues=[
        ["FW:123 SerialNo:serial_v"],
        ["FW:123 SerialNog:serial_v HW:123"],
        [""],
    ],
)
def test_parse_hs_device_information_failure(input_str: str) -> None:
    """Test parse device information."""
    with pytest.raises(utils.ParseError):
        utils.parse_hs_device_information(input_str)


@pytest.mark.parametrize(
    argnames=["input_str", "expected_result"],
    argvalues=[
        ["T:none C:123.4", Temperature(target=None, current=123.4)],
        ["T:123.566 C:123.446", Temperature(target=123.57, current=123.45)],
        ["T:-123.566 C:-123.446", Temperature(target=-123.57, current=-123.45)],
        ["a:3 T:2. C:3 H:0 G:1", Temperature(target=2, current=3)],
        ["T:0 C:124", Temperature(target=0, current=124)],
        ["T:none C:124", Temperature(target=None, current=124)],
    ],
)
def test_parse_temperature_response_success(
    input_str: str, expected_result: Temperature
) -> None:
    """It should parse temperature response."""
    assert utils.parse_temperature_response(input_str, 2) == expected_result


@pytest.mark.parametrize(
    argnames=["input_str"],
    argvalues=[
        ["T:not_a_float C:123"],
        ["T:none C:none"],
        ["C:not_a_float T:123"],
        [""],
    ],
)
def test_parse_temperature_response_failure(input_str: str) -> None:
    """It should fail to parse temperature response."""
    with pytest.raises(utils.ParseError):
        utils.parse_temperature_response(input_str, 2)


@pytest.mark.parametrize(
    argnames=["input_str", "expected_result"],
    argvalues=[
        ["T:0 C:500", RPM(target=None, current=500)],
        ["T:25 C:24", RPM(target=25, current=24)],
        ["a:3 T:2. C:3 H:0 G:1", RPM(target=2, current=3)],
    ],
)
def test_parse_rpm_response_success(input_str: str, expected_result: RPM) -> None:
    """It should parse temperature response."""
    assert utils.parse_rpm_response(input_str) == expected_result


@pytest.mark.parametrize(
    argnames=["input_str"],
    argvalues=[
        ["T:not_an_int C:123"],
        ["T:none C:none"],
        ["C:not_an_int T:123"],
        [""],
    ],
)
def test_parse_rpm_response_failure(input_str: str) -> None:
    """It should fail to parse temperature response."""
    with pytest.raises(utils.ParseError):
        utils.parse_rpm_response(input_str)


@pytest.mark.parametrize(
    argnames=["input_str", "expected_result"],
    argvalues=[
        ["STATUS:IDLE_OPEN", HeaterShakerLabwareLatchStatus.IDLE_OPEN],
        ["STATUS:IDLE_CLOSED", HeaterShakerLabwareLatchStatus.IDLE_CLOSED],
        ["a:safa STATUS:IDLE_UNKNOWN", HeaterShakerLabwareLatchStatus.IDLE_UNKNOWN],
    ],
)
def test_parse_labware_latch_status_response_success(
    input_str: str, expected_result: HeaterShakerLabwareLatchStatus
) -> None:
    assert utils.parse_labware_latch_status_response(input_str) == expected_result


@pytest.mark.parametrize(
    argnames=["input_str"],
    argvalues=[
        [""],
        ["SSSSSS:IDLE_OPEN"],
        ["STATUS:Lord Humongous Sees You"],
        ["IDLE_OPEN"],
        ["STATUS"],
        [":"],
    ],
)
def test_parse_labware_latch_status_response_failure(input_str: str) -> None:
    with pytest.raises(utils.ParseError):
        utils.parse_labware_latch_status_response(input_str)


@pytest.mark.parametrize(
    argnames=["input_str", "expected_result"],
    argvalues=[
        ["T:none C:0 H:none", PlateTemperature(target=None, current=0, hold=None)],
        [
            "T:321.4 C:45 H:123.222",
            PlateTemperature(target=321.4, current=45, hold=123.22),
        ],
        [
            "T:-44.2442 C:-22.233 H:0",
            PlateTemperature(target=-44.24, current=-22.23, hold=0),
        ],
        ["a:3 T:2. C:3 H:0 G:1", PlateTemperature(target=2, current=3, hold=0)],
    ],
)
def test_parse_plate_temperature_response_success(
    input_str: str, expected_result: PlateTemperature
) -> None:
    """It should parse plate temperature response"""
    assert utils.parse_plate_temperature_response(input_str, 2) == expected_result


@pytest.mark.parametrize(
    argnames=["input_str"],
    argvalues=[
        ["T:not_a_float C:123 H:321"],
        ["C:not_a_float T:123 H:321"],
        ["H:not_a_float C:123 T:321"],
        [""],
    ],
)
def test_parse_plate_temperature_response_failure(input_str: str) -> None:
    """It should fail to parse plate temperature response"""
    with pytest.raises(utils.ParseError):
        utils.parse_plate_temperature_response(input_str, 2)


@pytest.mark.parametrize(
    argnames=["input_str", "min_length", "expected"],
    argvalues=[
        ["", 0, ""],
        ["", 1, "0"],
        ["abcd", 12, "616263640000"],
        ["abcd", 5, "61626364"],
    ],
)
def test_string_to_hex(input_str: str, min_length: int, expected: str) -> None:
    """It should convert string to hex with padding to reach min_length."""
    assert expected == utils.string_to_hex(val=input_str, min_length=min_length)



import importlib
from typing import TYPE_CHECKING, Any

import pytest

import opentrons
from opentrons.calibration_storage import (
    delete_robot_belt_attitude,
    delete_robot_deck_attitude,
    get_robot_belt_attitude,
    get_robot_deck_attitude,
    save_robot_belt_attitude,
    save_robot_deck_attitude,
)
from opentrons.calibration_storage.ot2.models.v1 import (
    DeckCalibrationModel as OT2DeckCalModel,
)

# needed for proper type checking unfortunately
from opentrons.calibration_storage.ot3.models.v1 import (
    BeltCalibrationModel as OT3BeltCalModel,
)

if TYPE_CHECKING:
    from opentrons_shared_data.deck.types import RobotModel


@pytest.fixture(autouse=True)
def reload_module(robot_model: "RobotModel") -> None:
    importlib.reload(opentrons.calibration_storage)


@pytest.fixture
def starting_ot2_calibration_data(ot_config_tempdir: Any) -> None:
    """Starting OT-2 deck calibration data fixture."""
    save_robot_deck_attitude(
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        "pip1",
        "mytiprack",
    )


@pytest.fixture
def starting_ot3_calibration_data(ot_config_tempdir: Any) -> None:
    """Starting OT-3 belt calibration data fixture."""
    save_robot_belt_attitude(
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        "pip1",
    )


def test_save_ot2_deck_attitude(ot_config_tempdir: Any) -> None:
    """Test saving an OT-2 deck attitude calibration."""
    assert get_robot_deck_attitude() is None
    save_robot_deck_attitude(
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        "pip1",
        "mytiprack",
    )
    assert get_robot_deck_attitude() is not None


def test_save_ot3_deck_attitude(ot_config_tempdir: Any) -> None:
    """Test saving an OT-3 belt attitude calibration."""
    assert get_robot_belt_attitude() is None
    save_robot_belt_attitude(
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        "pip1",
    )
    assert get_robot_belt_attitude() != {}


def test_get_ot2_deck_calibration(starting_ot2_calibration_data: Any) -> None:
    """Test ability to get an OT-2 deck calibration model."""
    robot_deck = get_robot_deck_attitude()
    assert robot_deck
    assert robot_deck == OT2DeckCalModel(
        attitude=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        last_modified=robot_deck.last_modified,
        source=robot_deck.source,
        pipette_calibrated_with="pip1",
        tiprack="mytiprack",
    )


def test_get_ot3_deck_calibration(starting_ot3_calibration_data: Any) -> None:
    """Test ability to get an OT-3 belt calibration model."""
    robot_belt = get_robot_belt_attitude()
    assert robot_belt
    assert robot_belt == OT3BeltCalModel(
        attitude=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        lastModified=robot_belt.lastModified,
        source=robot_belt.source,
        pipetteCalibratedWith="pip1",
    )


def test_delete_ot2_deck_calibration(starting_ot2_calibration_data: Any) -> None:
    """Test delete OT-2 deck calibration data."""
    robot_deck = get_robot_deck_attitude()
    assert robot_deck
    assert robot_deck.attitude == [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    delete_robot_deck_attitude()
    assert get_robot_deck_attitude() is None


def test_delete_ot3_deck_calibration(starting_ot3_calibration_data: Any) -> None:
    """Test delete OT-3 belt calibration data."""
    robot_belt = get_robot_belt_attitude()
    assert robot_belt
    assert robot_belt.attitude == [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    delete_robot_belt_attitude()
    assert get_robot_belt_attitude() is None



"""Test unsafe drop tip in place commands."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import OT3HardwareControlAPI
from opentrons.hardware_control.types import Axis
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.unsafe.unsafe_drop_tip_in_place import (
    UnsafeDropTipInPlaceImplementation,
    UnsafeDropTipInPlaceParams,
    UnsafeDropTipInPlaceResult,
)
from opentrons.protocol_engine.execution import TipHandler
from opentrons.protocol_engine.state.motion import PipetteLocationData
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    PipetteTipStateUpdate,
    PipetteUnknownFluidUpdate,
    StateUpdate,
)
from opentrons.types import MountType


@pytest.fixture
def mock_tip_handler(decoy: Decoy) -> TipHandler:
    """Get a mock TipHandler."""
    return decoy.mock(cls=TipHandler)


@pytest.mark.parametrize("channels", [1, 8, 96])
async def test_drop_tip_implementation(
    decoy: Decoy,
    mock_tip_handler: TipHandler,
    state_view: StateView,
    ot3_hardware_api: OT3HardwareControlAPI,
    channels: int,
) -> None:
    """A DropTip command should have an execution implementation."""
    subject = UnsafeDropTipInPlaceImplementation(
        tip_handler=mock_tip_handler,
        state_view=state_view,
        hardware_api=ot3_hardware_api,
    )

    params = UnsafeDropTipInPlaceParams(pipetteId="abc", homeAfter=False)
    decoy.when(state_view.motion.get_pipette_location(pipette_id="abc")).then_return(
        PipetteLocationData(mount=MountType.LEFT, critical_point=None)
    )
    decoy.when(state_view.pipettes.get_active_channels(params.pipetteId)).then_return(
        channels
    )

    result = await subject.execute(params)

    assert result == SuccessData(
        public=UnsafeDropTipInPlaceResult(),
        state_update=StateUpdate(
            pipette_tip_state=PipetteTipStateUpdate(
                pipette_id="abc",
                tip_geometry=None,
                tip_source=None,
            ),
            pipette_aspirated_fluid=PipetteUnknownFluidUpdate(pipette_id="abc"),
        ),
    )

    decoy.verify(
        await ot3_hardware_api.update_axis_position_estimations([Axis.P_L]),
        await mock_tip_handler.drop_tip(
            pipette_id="abc", home_after=False, ignore_plunger=(channels == 96)
        ),
        times=1,
    )



"""Tests for protocol geometry planning module."""

from opentrons.motion_planning import MoveType
from opentrons.protocol_api.labware import Labware
from opentrons.protocols.geometry.planning import get_move_type


def test_get_move_type_general(min_lw: Labware, min_lw2: Labware) -> None:
    """It should identify general moves."""
    from_loc = min_lw.wells()[0].top()
    to_loc = min_lw2.wells()[0].top()

    result = get_move_type(from_loc, to_loc)
    assert result == MoveType.GENERAL_ARC


def test_get_move_type_in_labware(min_lw: Labware) -> None:
    """It should identify general moves."""
    from_loc = min_lw.wells()[0].top()
    to_loc = min_lw.wells()[1].top()

    result = get_move_type(from_loc, to_loc)
    assert result == MoveType.IN_LABWARE_ARC


def test_get_move_type_in_well(min_lw: Labware) -> None:
    """It should identify general moves."""
    from_loc = min_lw.wells()[0].top()
    to_loc = min_lw.wells()[0].bottom()

    result = get_move_type(from_loc, to_loc)
    assert result == MoveType.DIRECT


def test_get_move_type_general_with_force_direct(
    min_lw: Labware, min_lw2: Labware
) -> None:
    """It should identify general moves."""
    from_loc = min_lw.wells()[0].top()
    to_loc = min_lw2.wells()[0].top()

    result = get_move_type(from_loc, to_loc, force_direct=True)
    assert result == MoveType.DIRECT



"""Test legacy module command mapping in an end-to-end environment."""

import textwrap
from datetime import datetime
from pathlib import Path

import pytest
from decoy import matchers

from opentrons.protocol_engine import ModuleModel, commands
from opentrons.protocol_reader import ProtocolReader
from opentrons.protocol_runner.create_simulating_orchestrator import (
    create_simulating_orchestrator,
)


@pytest.fixture()
def modules_legacy_python_protocol_file(tmp_path: Path) -> Path:
    """Get an on-disk, minimal fixture of legacy python protocol w/ modules."""
    path = tmp_path / "protocol-name.py"
    path.write_text(
        textwrap.dedent(
            """
            # my protocol
            metadata = {
                "apiLevel": "2.11",
            }
            def run(ctx):
                tip_rack = ctx.load_labware(
                    load_name="opentrons_96_tiprack_300ul",
                    location="2",
                )
                temp_module = ctx.load_module(
                    module_name="temperature module",
                    location="3"
                )
                mag_module = ctx.load_module(
                    module_name="magnetic module",
                    location="6"
                )
                thermocycler = ctx.load_module(
                    module_name="thermocycler"
                )
                heater_shaker = ctx.load_module(
                    module_name="heaterShakerModuleV1",
                    location="1"
                )
            """
        )
    )

    return path


async def test_runner_with_modules_in_legacy_python(
    modules_legacy_python_protocol_file: Path,
) -> None:
    """It should map legacy module commands."""
    protocol_reader = ProtocolReader()
    protocol_source = await protocol_reader.read_saved(
        files=[modules_legacy_python_protocol_file],
        directory=None,
    )

    subject = await create_simulating_orchestrator(
        robot_type="OT-2 Standard", protocol_config=protocol_source.config
    )
    result = await subject.run(deck_configuration=[], protocol_source=protocol_source)
    commands_result = [c for c in result.commands]
    await subject.finish()

    assert len(commands_result) == 6

    temp_module_result_captor = matchers.Captor()
    mag_module_result_captor = matchers.Captor()
    thermocycler_result_captor = matchers.Captor()
    heater_shaker_result_captor = matchers.Captor()

    assert commands_result[0] == commands.Home.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=commands.HomeParams(axes=None),
        notes=[],
        result=commands.HomeResult(),
    )
    assert commands_result[1] == commands.LoadLabware.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=matchers.Anything(),
        notes=[],
        result=matchers.Anything(),
    )

    assert commands_result[2] == commands.LoadModule.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=matchers.Anything(),
        notes=[],
        result=temp_module_result_captor,
    )

    assert commands_result[3] == commands.LoadModule.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=matchers.Anything(),
        notes=[],
        result=mag_module_result_captor,
    )

    assert commands_result[4] == commands.LoadModule.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=matchers.Anything(),
        notes=[],
        result=thermocycler_result_captor,
    )

    assert commands_result[5] == commands.LoadModule.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=matchers.Anything(),
        notes=[],
        result=heater_shaker_result_captor,
    )

    assert temp_module_result_captor.value.model == ModuleModel.TEMPERATURE_MODULE_V1
    assert mag_module_result_captor.value.model == ModuleModel.MAGNETIC_MODULE_V1
    assert thermocycler_result_captor.value.model == ModuleModel.THERMOCYCLER_MODULE_V1
    assert (
        heater_shaker_result_captor.value.model == ModuleModel.HEATER_SHAKER_MODULE_V1
    )



"""Protocol Engine types dealing with command preconditions."""

from pydantic import BaseModel, Field

from opentrons_shared_data.util import StrEnum


class PreconditionTypes(StrEnum):
    """Precondition types used for identification during state update."""

    IS_CAMERA_USED = "isCameraUsed"


class CommandPreconditions(BaseModel):
    """Preconditions of commands as described in protocol analysis."""

    isCameraUsed: bool = Field(
        default=False,
        description="Parameter to determine if a Camera is used in a protocol.",
    )



"""Tests for `opentrons.simulate`."""

from __future__ import annotations

import io
import json
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Generator, List, TextIO, cast

import pytest
from _pytest.fixtures import SubRequest

from opentrons_shared_data import get_shared_data_root, load_shared_data

from opentrons import protocols, simulate
from opentrons.protocol_api.core.engine import ENGINE_CORE_API_VERSION
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.execution.errors import ExceptionInProtocolError
from opentrons.protocols.types import ApiDeprecationError
from opentrons.util import entrypoint_util

if TYPE_CHECKING:
    from tests.opentrons.conftest import Bundle, Protocol


HERE = Path(__file__).parent


@pytest.fixture(params=[APIVersion(2, 0), ENGINE_CORE_API_VERSION])
def api_version(request: SubRequest) -> APIVersion:
    """Return an API version to test with.

    Newer API versions execute through Protocol Engine, and older API versions don't.
    The two codepaths are very different, so we need to test them both.
    """
    return cast(APIVersion, request.param)


@pytest.mark.parametrize(
    "protocol_file",
    [
        "testosaur_v2.py",
        pytest.param(
            "testosaur_v2_14.py",
            marks=pytest.mark.xfail(
                strict=True,
                reason=(
                    "We can't currently get bundle contents"
                    " from protocols run through Protocol Engine."
                ),
            ),
        ),
    ],
)
def test_simulate_function_apiv2_bundle(
    protocol: Protocol,
    protocol_file: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test that `simulate()` returns the expected bundle contents from a Python file."""
    monkeypatch.setenv("OT_API_FF_allowBundleCreation", "1")
    _, bundle_contents = simulate.simulate(protocol.filelike, protocol.filename)
    assert isinstance(bundle_contents, protocols.types.BundleContents)


@pytest.mark.parametrize("protocol_file", ["testosaur_v2.py", "testosaur_v2_14.py"])
def test_simulate_without_filename(protocol: Protocol, protocol_file: str) -> None:
    """`simulate()` should accept a protocol without a filename."""
    simulate.simulate(protocol.filelike)  # Should not raise.


@pytest.mark.parametrize(
    ("protocol_file", "expected_entries"),
    [
        (
            "testosaur_v2.py",
            [
                "Picking up tip from A1 of Opentrons OT-2 96 Tip Rack 1000 µL on 1",
                "Aspirating 100.0 uL from A1 of Corning 96 Well Plate 360 µL Flat on 2 at 500.0 uL/sec",
                "Dispensing 100.0 uL into B1 of Corning 96 Well Plate 360 µL Flat on 2 at 1000.0 uL/sec",
                "Dropping tip into H12 of Opentrons OT-2 96 Tip Rack 1000 µL on 1",
            ],
        ),
        (
            "testosaur_v2_14.py",
            [
                "Picking up tip from A1 of Opentrons OT-2 96 Tip Rack 1000 µL on slot 1",
                "Aspirating 100.0 uL from A1 of Corning 96 Well Plate 360 µL Flat on slot 2 at 500.0 uL/sec",
                "Dispensing 100.0 uL into B1 of Corning 96 Well Plate 360 µL Flat on slot 2 at 1000.0 uL/sec",
                "Dropping tip into H12 of Opentrons OT-2 96 Tip Rack 1000 µL on slot 1",
            ],
        ),
        (
            "ot2_drop_tip.py",
            [
                "Picking up tip from A1 of Opentrons OT-2 96 Tip Rack 300 µL on slot 5",
                "Dropping tip into Trash Bin on slot 12",
            ],
        ),
    ],
)
def test_simulate_function_apiv2_run_log(
    protocol: Protocol,
    protocol_file: str,
    expected_entries: List[str],
) -> None:
    """Test that `simulate()` returns the expected run log from a Python file."""
    run_log, _ = simulate.simulate(protocol.filelike, protocol.filename)
    assert [item["payload"]["text"] for item in run_log] == expected_entries


def test_simulate_function_json(
    get_json_protocol_fixture: Callable[[str, str, bool], str],
) -> None:
    """Test `simulate()` with a JSON file."""
    jp = get_json_protocol_fixture("3", "simple", False)
    filelike = io.StringIO(jp)
    runlog, bundle = simulate.simulate(filelike, "simple.json")
    assert bundle is None
    assert [item["payload"]["text"] for item in runlog] == [
        "Picking up tip from B1 of Opentrons 96 Tip Rack 10 µL on 1",
        "Aspirating 5.0 uL from A1 of Source Plate on 2 at 3.0 uL/sec",
        "Delaying for 0 minutes and 42.0 seconds",
        "Dispensing 4.5 uL into B1 of Dest Plate on 3 at 2.5 uL/sec",
        "Touching tip",
        "Blowing out into B1 of Dest Plate on 3 at 2.0 uL/sec",
        "Moving to 5",
        "Dropping tip into A1 of Trash on 12",
    ]


def test_simulate_function_bundle_apiv2(
    get_bundle_fixture: Callable[[str], Bundle],
) -> None:
    """Test `simulate()` with a .zip bundle."""
    bundle_fixture = get_bundle_fixture("simple_bundle")
    runlog, bundle = simulate.simulate(
        cast(TextIO, bundle_fixture["filelike"]),
        "simple_bundle.zip",
    )
    assert bundle is None
    assert [item["payload"]["text"] for item in runlog] == [
        "Transferring 1.0 from A1 of FAKE example labware on 1 to A4 of FAKE example labware on 1",
        "Picking up tip from A1 of Opentrons OT-2 96 Tip Rack 10 µL on 3",
        "Aspirating 1.0 uL from A1 of FAKE example labware on 1 at 5.0 uL/sec",
        "Dispensing 1.0 uL into A4 of FAKE example labware on 1 at 10.0 uL/sec",
        "Dropping tip into A1 of Opentrons Fixed Trash on 12",
        "Transferring 2.0 from A1 of FAKE example labware on 1 to A4 of FAKE example labware on 1",
        "Picking up tip from B1 of Opentrons OT-2 96 Tip Rack 10 µL on 3",
        "Aspirating 2.0 uL from A1 of FAKE example labware on 1 at 5.0 uL/sec",
        "Dispensing 2.0 uL into A4 of FAKE example labware on 1 at 10.0 uL/sec",
        "Dropping tip into A1 of Opentrons Fixed Trash on 12",
        "Transferring 3.0 from A1 of FAKE example labware on 1 to A4 of FAKE example labware on 1",
        "Picking up tip from C1 of Opentrons OT-2 96 Tip Rack 10 µL on 3",
        "Aspirating 3.0 uL from A1 of FAKE example labware on 1 at 5.0 uL/sec",
        "Dispensing 3.0 uL into A4 of FAKE example labware on 1 at 10.0 uL/sec",
        "Dropping tip into A1 of Opentrons Fixed Trash on 12",
    ]


@pytest.mark.parametrize("protocol_file", ["testosaur.py"])
def test_simulate_function_v1(protocol: Protocol, protocol_file: str) -> None:
    """Test `simulate()` with an obsolete Python file."""
    with pytest.raises(ApiDeprecationError):
        simulate.simulate(protocol.filelike, "testosaur.py")


@pytest.mark.parametrize("protocol_file", ["bug_aspirate_tip.py"])
def test_simulate_aspirate_tip(
    protocol: Protocol,
    protocol_file: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Integration test for https://github.com/opentrons/opentrons/issues/7552."""
    with pytest.raises(ExceptionInProtocolError):
        simulate.simulate(protocol.filelike, "bug_aspirate_tip.py")


class TestSimulatePythonLabware:
    """Tests for making sure simulate() handles custom labware correctly for Python files."""

    LW_DIR = get_shared_data_root() / "labware" / "fixtures" / "2"
    LW_LOAD_NAME = "fixture_12_trough"
    LW_NAMESPACE = "fixture"

    @pytest.fixture
    def protocol_path(self, tmp_path: Path, api_version: APIVersion) -> Path:
        """Return a path to a Python protocol file that loads a custom labware."""
        path = tmp_path / "protocol.py"
        protocol_source = textwrap.dedent(
            f"""\
            metadata = {{"apiLevel": "{api_version}"}}
            def run(protocol):
                protocol.load_labware(
                    load_name="{self.LW_LOAD_NAME}",
                    location=1,
                    namespace="{self.LW_NAMESPACE}",
                )
            """
        )
        path.write_text(protocol_source)
        return path

    @pytest.fixture
    def file_name(self, protocol_path: Path) -> str:
        """Return the file name of the Python protocol file."""
        return protocol_path.name

    @pytest.fixture
    def protocol_filelike(self, protocol_path: Path) -> Generator[TextIO, None, None]:
        """Return the Python protocol file opened as a stream."""
        with open(protocol_path) as file:
            yield file

    @staticmethod
    def test_default_no_custom_labware(
        protocol_filelike: TextIO, file_name: str
    ) -> None:
        """By default, no custom labware should be available."""
        with pytest.raises(Exception, match="Labware .+ not found"):
            simulate.simulate(protocol_file=protocol_filelike, file_name=file_name)

    def test_custom_labware_paths(
        self, protocol_filelike: TextIO, file_name: str
    ) -> None:
        """Providing custom_labware_paths should make those labware available."""
        simulate.simulate(
            protocol_file=protocol_filelike,
            file_name=file_name,
            custom_labware_paths=[str(self.LW_DIR)],
        )

    def test_jupyter(
        self,
        protocol_filelike: TextIO,
        file_name: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Putting labware in the Jupyter directory should make it available."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", self.LW_DIR
        )
        simulate.simulate(protocol_file=protocol_filelike, file_name=file_name)

    @pytest.mark.xfail(
        strict=True, raises=pytest.fail.Exception
    )  # TODO(mm, 2023-07-14): Fix this bug.
    def test_jupyter_override(
        self,
        protocol_filelike: TextIO,
        file_name: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Passing any custom_labware_paths should prevent searching the Jupyter directory."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", self.LW_DIR
        )
        with pytest.raises(Exception, match="Labware .+ not found"):
            simulate.simulate(
                protocol_file=protocol_filelike,
                file_name=file_name,
                custom_labware_paths=[],
            )

    @staticmethod
    def test_jupyter_not_on_filesystem(
        protocol_filelike: TextIO,
        file_name: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """It should tolerate the Jupyter labware directory not existing on the filesystem."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", HERE / "nosuchdirectory"
        )
        with pytest.raises(Exception, match="Labware .+ not found"):
            simulate.simulate(protocol_file=protocol_filelike, file_name=file_name)


def test_get_protocol_api_usable_without_homing(api_version: APIVersion) -> None:
    """You should be able to move the simulated hardware without having to home explicitly.

    https://opentrons.atlassian.net/browse/RQA-1801
    """
    protocol = simulate.get_protocol_api(api_version)
    pipette = protocol.load_instrument("p300_single_gen2", mount="left")
    tip_rack = protocol.load_labware("opentrons_96_tiprack_300ul", 1)
    pipette.pick_up_tip(tip_rack["A1"])  # Should not raise.


def test_liquid_probe_get_protocol_api() -> None:
    """Covers `simulate.get_protocol_api()`-specific issues with liquid probes.

    See https://opentrons.atlassian.net/browse/EXEC-646.
    """
    protocol = simulate.get_protocol_api(version="2.20", robot_type="Flex")
    pipette = protocol.load_instrument("flex_1channel_1000", mount="left")
    tip_rack = protocol.load_labware("opentrons_flex_96_tiprack_1000ul", "A1")
    well_plate = protocol.load_labware(
        "opentrons_96_wellplate_200ul_pcr_full_skirt", "A2"
    )
    pipette.pick_up_tip(tip_rack["A1"])
    pipette.require_liquid_presence(well_plate["A1"])  # Should not raise MustHomeError.


def test_liquid_probe_simulate_file() -> None:
    """Covers `opentrons_simulate`-specific issues with liquid probes.

    See https://opentrons.atlassian.net/browse/EXEC-646.
    """
    protocol_contents = textwrap.dedent(
        """\
        requirements = {"robotType": "Flex", "apiLevel": "2.20"}
        def run(protocol):
            pipette = protocol.load_instrument("flex_1channel_1000", mount="left")
            tip_rack = protocol.load_labware("opentrons_flex_96_tiprack_1000ul", "A1")
            well_plate = protocol.load_labware(
                "opentrons_96_wellplate_200ul_pcr_full_skirt", "A2"
            )
            pipette.pick_up_tip(tip_rack["A1"])
            pipette.require_liquid_presence(well_plate["A1"])
        """
    )
    protocol_contents_stream = io.StringIO(protocol_contents)
    simulate.simulate(
        protocol_file=protocol_contents_stream
    )  # Should not raise MustHomeError.


class TestGetProtocolAPILabware:
    """Tests for making sure get_protocol_api() handles extra labware correctly."""

    LW_FIXTURE_DIR = Path("labware/fixtures/2")
    LW_LOAD_NAME = "fixture_12_trough"
    LW_NAMESPACE = "fixture"

    def test_default_no_extra_labware(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """By default, no extra labware should be available."""
        context = simulate.get_protocol_api(api_version)
        with pytest.raises(Exception, match="Labware .+ not found"):
            context.load_labware(
                load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
            )

    def test_extra_labware(self, api_version: APIVersion) -> None:
        """Providing extra_labware should make that labware available."""
        explicit_extra_lw = {
            self.LW_LOAD_NAME: json.loads(
                load_shared_data(self.LW_FIXTURE_DIR / f"{self.LW_LOAD_NAME}.json")
            )
        }
        context = simulate.get_protocol_api(
            api_version, extra_labware=explicit_extra_lw
        )
        assert context.load_labware(
            load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
        )

    def test_jupyter(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Putting labware in the Jupyter directory should make it available."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util,
            "JUPYTER_NOTEBOOK_LABWARE_DIR",
            get_shared_data_root() / self.LW_FIXTURE_DIR,
        )
        context = simulate.get_protocol_api(api_version)
        assert context.load_labware(
            load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
        )

    def test_jupyter_override(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Passing any extra_labware should prevent searching the Jupyter directory."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util,
            "JUPYTER_NOTEBOOK_LABWARE_DIR",
            get_shared_data_root() / self.LW_FIXTURE_DIR,
        )
        context = simulate.get_protocol_api(api_version, extra_labware={})
        with pytest.raises(Exception, match="Labware .+ not found"):
            context.load_labware(
                load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
            )

    def test_jupyter_not_on_filesystem(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """It should tolerate the Jupyter labware directory not existing on the filesystem."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", HERE / "nosuchdirectory"
        )
        with_nonexistent_jupyter_extra_labware = simulate.get_protocol_api(api_version)
        with pytest.raises(Exception, match="Labware .+ not found"):
            with_nonexistent_jupyter_extra_labware.load_labware(
                load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
            )



"""Test update-position-estimator commands."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import OT3HardwareControlAPI
from opentrons.hardware_control.types import Axis
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.unsafe.unsafe_ungrip_labware import (
    UnsafeUngripLabwareImplementation,
    UnsafeUngripLabwareParams,
    UnsafeUngripLabwareResult,
)
from opentrons.protocol_engine.errors.exceptions import GripperNotAttachedError


async def test_ungrip_labware_implementation(
    decoy: Decoy, ot3_hardware_api: OT3HardwareControlAPI
) -> None:
    """Test UngripLabware command execution."""
    subject = UnsafeUngripLabwareImplementation(hardware_api=ot3_hardware_api)

    decoy.when(ot3_hardware_api.has_gripper()).then_return(True)

    result = await subject.execute(params=UnsafeUngripLabwareParams())

    assert result == SuccessData(public=UnsafeUngripLabwareResult())

    decoy.verify(
        await ot3_hardware_api.home([Axis.G]),
    )


async def test_ungrip_labware_implementation_raises_no_gripper_attached(
    decoy: Decoy, ot3_hardware_api: OT3HardwareControlAPI
) -> None:
    """Test UngripLabware command execution."""
    subject = UnsafeUngripLabwareImplementation(hardware_api=ot3_hardware_api)

    decoy.when(ot3_hardware_api.has_gripper()).then_return(False)
    with pytest.raises(GripperNotAttachedError):
        await subject.execute(params=UnsafeUngripLabwareParams())



"""Test drop tip commands."""

from datetime import datetime
from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition2,
    Parameters2,
)

from opentrons.hardware_control.types import TipScrapeType
from opentrons.protocol_engine import (
    DeckPoint,
    DropTipWellLocation,
    DropTipWellOrigin,
    WellLocation,
    WellOffset,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.drop_tip import (
    DropTipImplementation,
    DropTipParams,
    DropTipResult,
)
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.commands.pipetting_common import (
    TipPhysicallyAttachedError,
)
from opentrons.protocol_engine.errors.exceptions import TipAttachedError
from opentrons.protocol_engine.execution import MovementHandler, TipHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId, TipRackWellState
from opentrons.types import Point


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def mock_movement_handler(decoy: Decoy) -> MovementHandler:
    """Get a mock MovementHandler."""
    return decoy.mock(cls=MovementHandler)


@pytest.fixture
def mock_tip_handler(decoy: Decoy) -> TipHandler:
    """Get a mock TipHandler."""
    return decoy.mock(cls=TipHandler)


@pytest.fixture
def mock_model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mock ModelUtils."""
    return decoy.mock(cls=ModelUtils)


def test_drop_tip_params_defaults() -> None:
    """A drop tip should use a `WellOrigin.DROP_TIP` by default."""
    default_params = DropTipParams.model_validate(
        {"pipetteId": "abc", "labwareId": "def", "wellName": "ghj"}
    )

    assert default_params.wellLocation == DropTipWellLocation(
        origin=DropTipWellOrigin.DEFAULT, offset=WellOffset(x=0, y=0, z=0)
    )


def test_drop_tip_params_default_origin() -> None:
    """A drop tip should drop a `WellOrigin.DROP_TIP` by default even if an offset is given."""
    default_params = DropTipParams.model_validate(
        {
            "pipetteId": "abc",
            "labwareId": "def",
            "wellName": "ghj",
            "wellLocation": {"offset": {"x": 1, "y": 2, "z": 3}},
        }
    )

    assert default_params.wellLocation == DropTipWellLocation(
        origin=DropTipWellOrigin.DEFAULT, offset=WellOffset(x=1, y=2, z=3)
    )


async def test_drop_tip_implementation(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_movement_handler: MovementHandler,
    mock_tip_handler: TipHandler,
    mock_model_utils: ModelUtils,
) -> None:
    """A DropTip command should have an execution implementation."""
    subject = DropTipImplementation(
        state_view=mock_state_view,
        movement=mock_movement_handler,
        tip_handler=mock_tip_handler,
        model_utils=mock_model_utils,
    )

    params = DropTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=DropTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        homeAfter=True,
    )

    decoy.when(
        mock_state_view.geometry.get_checked_tip_drop_location(
            pipette_id="abc",
            labware_id="123",
            well_location=DropTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        )
    ).then_return(WellLocation(offset=WellOffset(x=4, y=5, z=6)))

    decoy.when(mock_state_view.labware.get_definition("123")).then_return(
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            parameters=Parameters2.model_construct(isTiprack=True),  # type: ignore[call-arg]
        )
    )
    decoy.when(mock_state_view.pipettes.get_nozzle_configuration("abc")).then_return(
        sentinel.nozzle_configuration
    )
    decoy.when(
        mock_state_view.tips.compute_tips_to_mark_as_used_or_empty(
            "123", "A3", sentinel.nozzle_configuration
        )
    ).then_return(sentinel.tips_to_mark_as_used)

    decoy.when(
        await mock_movement_handler.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=4, y=5, z=6)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=111, y=222, z=333))

    result = await subject.execute(params)

    assert result == SuccessData(
        public=DropTipResult(position=DeckPoint(x=111, y=222, z=333)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(
                    labware_id="123",
                    well_name="A3",
                ),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
            pipette_tip_state=update_types.PipetteTipStateUpdate(
                pipette_id="abc",
                tip_geometry=None,
                tip_source=None,
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="abc"
            ),
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.USED,
                labware_id="123",
                well_names=sentinel.tips_to_mark_as_used,
            ),
        ),
    )

    decoy.verify(
        await mock_tip_handler.drop_tip(
            pipette_id="abc", home_after=True, scrape_type=TipScrapeType.NONE
        ),
        times=1,
    )


async def test_drop_tip_with_alternating_locations(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_movement_handler: MovementHandler,
    mock_tip_handler: TipHandler,
    mock_model_utils: ModelUtils,
) -> None:
    """It should drop tip at random location within the labware every time."""
    subject = DropTipImplementation(
        state_view=mock_state_view,
        movement=mock_movement_handler,
        tip_handler=mock_tip_handler,
        model_utils=mock_model_utils,
    )
    params = DropTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=DropTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        homeAfter=True,
        alternateDropLocation=True,
    )
    drop_location = DropTipWellLocation(
        origin=DropTipWellOrigin.DEFAULT, offset=WellOffset(x=10, y=20, z=30)
    )
    decoy.when(
        mock_state_view.geometry.get_next_tip_drop_location(
            labware_id="123", well_name="A3", pipette_id="abc"
        )
    ).then_return(drop_location)

    decoy.when(
        mock_state_view.geometry.get_checked_tip_drop_location(
            pipette_id="abc",
            labware_id="123",
            well_location=drop_location,
        )
    ).then_return(WellLocation(offset=WellOffset(x=4, y=5, z=6)))

    decoy.when(mock_state_view.labware.get_definition("123")).then_return(
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            parameters=Parameters2.model_construct(isTiprack=False),  # type: ignore[call-arg]
        )
    )

    decoy.when(
        await mock_movement_handler.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=4, y=5, z=6)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=111, y=222, z=333))

    result = await subject.execute(params)
    assert result == SuccessData(
        public=DropTipResult(position=DeckPoint(x=111, y=222, z=333)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(
                    labware_id="123",
                    well_name="A3",
                ),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
            pipette_tip_state=update_types.PipetteTipStateUpdate(
                pipette_id="abc",
                tip_geometry=None,
                tip_source=None,
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="abc"
            ),
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.USED,
                labware_id="123",
                well_names=[],
            ),
        ),
    )


async def test_tip_attached_error(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_movement_handler: MovementHandler,
    mock_tip_handler: TipHandler,
    mock_model_utils: ModelUtils,
) -> None:
    """A DropTip command should have an execution implementation."""
    subject = DropTipImplementation(
        state_view=mock_state_view,
        movement=mock_movement_handler,
        tip_handler=mock_tip_handler,
        model_utils=mock_model_utils,
    )

    params = DropTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=DropTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        scrape_tips=False,
    )

    decoy.when(
        mock_state_view.geometry.get_checked_tip_drop_location(
            pipette_id="abc",
            labware_id="123",
            well_location=DropTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        )
    ).then_return(WellLocation(offset=WellOffset(x=4, y=5, z=6)))

    decoy.when(mock_state_view.labware.get_definition("123")).then_return(
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            parameters=Parameters2.model_construct(isTiprack=True),  # type: ignore[call-arg]
        )
    )
    decoy.when(mock_state_view.pipettes.get_nozzle_configuration("abc")).then_return(
        sentinel.nozzle_configuration
    )
    decoy.when(
        mock_state_view.tips.compute_tips_to_mark_as_used_or_empty(
            "123", "A3", sentinel.nozzle_configuration
        )
    ).then_return(sentinel.tips_to_mark_as_used)

    decoy.when(
        await mock_movement_handler.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=4, y=5, z=6)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=111, y=222, z=333))
    decoy.when(
        await mock_tip_handler.drop_tip(  # type: ignore[func-returns-value]
            pipette_id="abc", home_after=None, scrape_type=TipScrapeType.NONE
        )
    ).then_raise(TipAttachedError("Egads!"))

    decoy.when(mock_model_utils.generate_id()).then_return("error-id")
    decoy.when(mock_model_utils.get_timestamp()).then_return(
        datetime(year=1, month=2, day=3)
    )

    result = await subject.execute(params)

    assert result == DefinedErrorData(
        public=TipPhysicallyAttachedError.model_construct(
            id="error-id",
            createdAt=datetime(year=1, month=2, day=3),
            wrappedErrors=[matchers.Anything()],
            errorInfo={"retryLocation": (111, 222, 333)},
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(
                    labware_id="123",
                    well_name="A3",
                ),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="abc"
            ),
        ),
        state_update_if_false_positive=update_types.StateUpdate(
            pipette_tip_state=update_types.PipetteTipStateUpdate(
                pipette_id="abc",
                tip_geometry=None,
                tip_source=None,
            ),
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(
                    labware_id="123",
                    well_name="A3",
                ),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.USED,
                labware_id="123",
                well_names=sentinel.tips_to_mark_as_used,
            ),
        ),
    )


async def test_stall_error(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_movement_handler: MovementHandler,
    mock_tip_handler: TipHandler,
    mock_model_utils: ModelUtils,
) -> None:
    """A DropTip command should have an execution implementation."""
    subject = DropTipImplementation(
        state_view=mock_state_view,
        movement=mock_movement_handler,
        tip_handler=mock_tip_handler,
        model_utils=mock_model_utils,
    )

    params = DropTipParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=DropTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
    )

    decoy.when(
        mock_state_view.geometry.get_checked_tip_drop_location(
            pipette_id="abc",
            labware_id="123",
            well_location=DropTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        )
    ).then_return(WellLocation(offset=WellOffset(x=4, y=5, z=6)))

    decoy.when(mock_state_view.labware.get_definition("123")).then_return(
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            parameters=Parameters2.model_construct(isTiprack=False),  # type: ignore[call-arg]
        )
    )

    decoy.when(
        await mock_movement_handler.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=4, y=5, z=6)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_raise(StallOrCollisionDetectedError())

    decoy.when(mock_model_utils.generate_id()).then_return("error-id")
    decoy.when(mock_model_utils.get_timestamp()).then_return(
        datetime(year=1, month=2, day=3)
    )

    result = await subject.execute(params)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id="error-id",
            createdAt=datetime(year=1, month=2, day=3),
            wrappedErrors=[matchers.Anything()],
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
        ),
    )



"""Tests for heater_shaker_movement_flagger."""

from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager, NamedTuple

import pytest
from decoy import Decoy

from opentrons_shared_data.robot.types import RobotType

from opentrons.drivers.types import HeaterShakerLabwareLatchStatus
from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.modules.heater_shaker import (
    HeaterShaker as HardwareHeaterShaker,
)
from opentrons.protocol_engine.errors import (
    HeaterShakerLabwareLatchNotOpenError,
    HeaterShakerLabwareLatchStatusUnknown,
    PipetteMovementRestrictedByHeaterShakerError,
    WrongModuleTypeError,
)
from opentrons.protocol_engine.execution.heater_shaker_movement_flagger import (
    HeaterShakerMovementFlagger,
)
from opentrons.protocol_engine.state.module_substates.heater_shaker_module_substate import (
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    HeaterShakerLatchStatus,
    HeaterShakerMovementRestrictors,
    ModuleLocation,
)
from opentrons.types import DeckSlotName


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mock in the shape of a HardwareAPI."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mock in the shape of a StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def subject(
    state_store: StateStore,
    hardware_api: HardwareAPI,
) -> HeaterShakerMovementFlagger:
    """Return a h/s movement flagger initialized with mocked-out dependencies."""
    return HeaterShakerMovementFlagger(
        state_store=state_store,
        hardware_api=hardware_api,
    )


@pytest.mark.parametrize(
    argnames=["destination_slot", "expected_raise"],
    argvalues=[
        [
            4,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # east
        [
            6,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # west
        [
            8,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # north
        [
            2,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # south
        [
            5,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # h/s
        [1, does_not_raise()],  # non-adjacent
    ],
)
async def test_raises_when_moving_to_restricted_slots_while_shaking(
    subject: HeaterShakerMovementFlagger,
    destination_slot: int,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise if restricted movement around a heater-shaker is attempted while module is shaking."""
    heater_shaker_data = [
        HeaterShakerMovementRestrictors(
            plate_shaking=True, latch_status=HeaterShakerLatchStatus.CLOSED, deck_slot=5
        )
    ]

    with expected_raise:
        subject.raise_if_movement_restricted(
            hs_movement_restrictors=heater_shaker_data,
            destination_slot=destination_slot,
            is_multi_channel=False,
            destination_is_tip_rack=False,
        )


@pytest.mark.parametrize(
    argnames=["robot_type", "destination_slot", "expected_raise"],
    argvalues=[
        [
            "OT-2 Standard",
            4,
            pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="latch"),
        ],  # east on OT2
        [
            "OT-2 Standard",
            6,
            pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="latch"),
        ],  # west on OT2
        [
            "OT-2 Standard",
            5,
            pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="latch"),
        ],  # to h/s on OT2
        ["OT-2 Standard", 8, does_not_raise()],  # north on OT2
        ["OT-2 Standard", 2, does_not_raise()],  # south on OT2
        ["OT-2 Standard", 3, does_not_raise()],  # non-adjacent
        [
            "OT-3 Standard",
            5,
            pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="latch"),
        ],  # to h/s on OT3
        ["OT-3 Standard", 4, does_not_raise()],  # east on OT3
        ["OT-3 Standard", 6, does_not_raise()],  # west on OT3
        ["OT-3 Standard", 8, does_not_raise()],  # north on OT3
        ["OT-3 Standard", 2, does_not_raise()],  # south on OT3
    ],
)
async def test_raises_when_moving_to_restricted_slots_while_latch_open2(
    decoy: Decoy,
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    robot_type: RobotType,
    destination_slot: int,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise if restricted movement around a heater-shaker is attempted while latch is open."""
    heater_shaker_data = [
        HeaterShakerMovementRestrictors(
            plate_shaking=False, latch_status=HeaterShakerLatchStatus.OPEN, deck_slot=5
        )
    ]
    decoy.when(state_store.config.robot_type).then_return(robot_type)
    with expected_raise:
        subject.raise_if_movement_restricted(
            hs_movement_restrictors=heater_shaker_data,
            destination_slot=destination_slot,
            is_multi_channel=False,
            destination_is_tip_rack=False,
        )


@pytest.mark.parametrize(
    argnames=["destination_slot", "is_tiprack", "expected_raise"],
    argvalues=[
        [
            4,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="left or right"
            ),
        ],  # east
        [
            6,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="left or right"
            ),
        ],  # west
        [
            8,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError,
                match="non-tip-rack labware",
            ),
        ],  # north, non-tiprack
        [
            2,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError,
                match="non-tip-rack labware",
            ),
        ],  # south, non-tiprack
        [8, True, does_not_raise()],  # north, tiprack
        [2, True, does_not_raise()],  # south, tiprack
        [5, False, does_not_raise()],  # h/s
        [7, False, does_not_raise()],  # non-adjacent
    ],
)
def test_raises_on_restricted_movement_with_multi_channel(
    decoy: Decoy,
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    destination_slot: int,
    is_tiprack: bool,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise if restricted movement around a heater-shaker is attempted with a multi-channel pipette."""
    heater_shaker_data = [
        HeaterShakerMovementRestrictors(
            plate_shaking=False,
            latch_status=HeaterShakerLatchStatus.CLOSED,
            deck_slot=5,
        )
    ]
    decoy.when(state_store.config.robot_type).then_return("OT-2 Standard")
    with expected_raise:
        subject.raise_if_movement_restricted(
            hs_movement_restrictors=heater_shaker_data,
            destination_slot=destination_slot,
            is_multi_channel=True,
            destination_is_tip_rack=is_tiprack,
        )


def test_does_not_raise_on_movement_with_multi_channel_on_ot3(
    decoy: Decoy,
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
) -> None:
    """It should not raise when pipetting with multichannel around a H/S on OT3."""
    heater_shaker_data = [
        HeaterShakerMovementRestrictors(
            plate_shaking=False,
            latch_status=HeaterShakerLatchStatus.CLOSED,
            deck_slot=5,
        )
    ]
    decoy.when(state_store.config.robot_type).then_return("OT-3 Standard")
    subject.raise_if_movement_restricted(
        hs_movement_restrictors=heater_shaker_data,
        destination_slot=6,
        is_multi_channel=True,
        destination_is_tip_rack=False,
    )


@pytest.mark.parametrize(
    argnames=["destination_slot"],
    argvalues=[
        [4],  # east
        [6],  # west
        [5],  # h/s
        [8],  # north
        [2],  # south
        [9],  # non-adjacent
    ],
)
async def test_does_not_raise_when_idle_and_latch_closed(
    subject: HeaterShakerMovementFlagger,
    destination_slot: int,
) -> None:
    """It should not raise if single channel pipette moves anywhere near heater-shaker when idle and latch closed."""
    heater_shaker_data = [
        HeaterShakerMovementRestrictors(
            plate_shaking=False,
            latch_status=HeaterShakerLatchStatus.CLOSED,
            deck_slot=5,
        )
    ]

    with does_not_raise():
        subject.raise_if_movement_restricted(
            hs_movement_restrictors=heater_shaker_data,
            destination_slot=destination_slot,
            is_multi_channel=False,
            destination_is_tip_rack=False,
        )


@pytest.mark.parametrize(
    argnames=["latch_status", "expected_raise_cm"],
    argvalues=[
        (
            HeaterShakerLatchStatus.CLOSED,
            pytest.raises(HeaterShakerLabwareLatchNotOpenError),
        ),
        (
            HeaterShakerLatchStatus.UNKNOWN,
            pytest.raises(HeaterShakerLabwareLatchStatusUnknown),
        ),
    ],
)
async def test_raises_depending_on_heater_shaker_substate_latch_status(
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    decoy: Decoy,
    latch_status: HeaterShakerLatchStatus,
    expected_raise_cm: ContextManager[Any],
) -> None:
    """It should flag movement depending on engine's h/s state."""
    decoy.when(
        state_store.modules.get_heater_shaker_module_substate(module_id="module-id")
    ).then_return(
        HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=latch_status,
            is_plate_shaking=False,
            plate_target_temperature=None,
        )
    )

    with expected_raise_cm:
        await subject.raise_if_labware_latched_on_heater_shaker(
            labware_parent=ModuleLocation(moduleId="module-id")
        )


class LatchStatusAndRaiseSpec(NamedTuple):
    """Test parametrization data.

    A H/S labware latch status, and what we expect the subject to raise when a h/s
    has that latch status.
    """

    latch_status: HeaterShakerLabwareLatchStatus
    expected_raise_cm: ContextManager[Any]


@pytest.mark.parametrize(
    LatchStatusAndRaiseSpec._fields,
    [
        LatchStatusAndRaiseSpec(
            latch_status=HeaterShakerLabwareLatchStatus.IDLE_UNKNOWN,
            expected_raise_cm=pytest.raises(HeaterShakerLabwareLatchNotOpenError),
        ),
        LatchStatusAndRaiseSpec(
            latch_status=HeaterShakerLabwareLatchStatus.IDLE_CLOSED,
            expected_raise_cm=pytest.raises(HeaterShakerLabwareLatchNotOpenError),
        ),
        LatchStatusAndRaiseSpec(
            latch_status=HeaterShakerLabwareLatchStatus.CLOSING,
            expected_raise_cm=pytest.raises(HeaterShakerLabwareLatchNotOpenError),
        ),
        LatchStatusAndRaiseSpec(
            latch_status=HeaterShakerLabwareLatchStatus.OPENING,
            expected_raise_cm=pytest.raises(HeaterShakerLabwareLatchNotOpenError),
        ),
        LatchStatusAndRaiseSpec(
            latch_status=HeaterShakerLabwareLatchStatus.UNKNOWN,
            expected_raise_cm=pytest.raises(HeaterShakerLabwareLatchNotOpenError),
        ),
        LatchStatusAndRaiseSpec(
            latch_status=HeaterShakerLabwareLatchStatus.IDLE_OPEN,
            expected_raise_cm=does_not_raise(),
        ),
    ],
)
async def test_raises_depending_on_heater_shaker_latch_status(
    decoy: Decoy,
    latch_status: HeaterShakerLabwareLatchStatus,
    expected_raise_cm: ContextManager[Any],
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
) -> None:
    """When on a H/S, it should raise if the latch isn't open."""
    decoy.when(
        state_store.modules.get_heater_shaker_module_substate(module_id="module-id")
    ).then_return(
        HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.OPEN,
            is_plate_shaking=False,
            plate_target_temperature=None,
        )
    )

    decoy.when(state_store.config.use_virtual_modules).then_return(False)
    decoy.when(
        state_store.modules.get_serial_number(module_id="module-id")
    ).then_return("module-serial")

    heater_shaker = decoy.mock(cls=HardwareHeaterShaker)
    decoy.when(hardware_api.attached_modules).then_return([heater_shaker])
    decoy.when(heater_shaker.device_info).then_return({"serial": "module-serial"})
    decoy.when(heater_shaker.labware_latch_status).then_return(latch_status)

    with expected_raise_cm:
        await subject.raise_if_labware_latched_on_heater_shaker(
            labware_parent=ModuleLocation(moduleId="module-id"),
        )


async def test_raises_if_hardware_module_has_gone_missing(
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    decoy: Decoy,
) -> None:
    """It should raise if the hardware module can't be found by its serial no."""
    decoy.when(
        state_store.modules.get_heater_shaker_module_substate(module_id="module-id")
    ).then_return(
        HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.OPEN,
            is_plate_shaking=False,
            plate_target_temperature=None,
        ),
    )

    decoy.when(state_store.config.use_virtual_modules).then_return(False)
    decoy.when(
        state_store.modules.get_serial_number(module_id="module-id")
    ).then_return("module-serial")
    decoy.when(hardware_api.attached_modules).then_return([])

    with pytest.raises(HeaterShakerLabwareLatchNotOpenError):
        await subject.raise_if_labware_latched_on_heater_shaker(
            labware_parent=ModuleLocation(moduleId="module-id"),
        )


async def test_passes_if_virtual_module_latch_open(
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    decoy: Decoy,
) -> None:
    """It shouldn't raise if a virtual module in analysis has latch open."""
    decoy.when(
        state_store.modules.get_heater_shaker_module_substate(module_id="module-id")
    ).then_return(
        HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.OPEN,
            is_plate_shaking=False,
            plate_target_temperature=None,
        ),
    )
    decoy.when(state_store.config.use_virtual_modules).then_return(True)
    await subject.raise_if_labware_latched_on_heater_shaker(
        labware_parent=ModuleLocation(moduleId="module-id")
    )


async def test_passes_if_labware_on_non_heater_shaker_module(
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    decoy: Decoy,
) -> None:
    """It shouldn't raise if the labware is on a module other than a Heater-Shaker."""
    decoy.when(
        state_store.modules.get_heater_shaker_module_substate(module_id="module-id")
    ).then_raise(WrongModuleTypeError("Woops"))
    await subject.raise_if_labware_latched_on_heater_shaker(
        ModuleLocation(moduleId="module-id")
    )


async def test_passes_if_labware_not_on_any_module(
    subject: HeaterShakerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    decoy: Decoy,
) -> None:
    """It shouldn't raise if the labware isn't on a module."""
    await subject.raise_if_labware_latched_on_heater_shaker(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )



from typing import List

import pytest
from numpy import float64 as f64

from opentrons_hardware.firmware_bindings.constants import NodeId
from opentrons_hardware.hardware_control.motion_planning import Move

from opentrons.config import defaults_ot3
from opentrons.config import types as conf_types
from opentrons.hardware_control.backends import ot3utils
from opentrons.hardware_control.types import Axis, OT3AxisKind, OT3Mount


def test_create_step() -> None:
    origin = {
        Axis.X: 0,
        Axis.Y: 0,
        Axis.Z_L: 0,
        Axis.Z_R: 0,
        Axis.P_L: 0,
        Axis.P_R: 0,
    }
    moves = [Move.build_dummy([Axis.X, Axis.Y, Axis.Z_L, Axis.Z_R, Axis.P_L])]
    for block in moves[0].blocks:
        block.distance = f64(25.0)
        block.time = f64(1.0)
        block.initial_speed = f64(25.0)
        block.acceleration = f64(0.0)
        block.final_speed = f64(25.0)
    present_nodes = [NodeId.gantry_x, NodeId.gantry_y, NodeId.head_l]
    move_group, final_pos = ot3utils.create_move_group(
        origin=origin,
        moves=moves,
        present_nodes=present_nodes,
    )
    assert len(move_group) == 3
    for step in move_group:
        assert set(present_nodes) == set(step.keys())


def test_add_group_delay() -> None:
    origin = {
        Axis.X: 0,
        Axis.Y: 0,
        Axis.Z_L: 0,
        Axis.Z_R: 0,
        Axis.P_L: 0,
        Axis.P_R: 0,
    }
    moves = [Move.build_dummy([Axis.X, Axis.Y, Axis.Z_L, Axis.Z_R, Axis.P_L])]
    for block in moves[0].blocks:
        block.distance = f64(25.0)
        block.time = f64(1.0)
        block.initial_speed = f64(25.0)
        block.acceleration = f64(0.0)
        block.final_speed = f64(25.0)
    present_nodes = [NodeId.gantry_x, NodeId.gantry_y, NodeId.head_l]
    move_group, final_pos = ot3utils.create_move_group(
        origin=origin,
        moves=moves,
        present_nodes=present_nodes,
    )
    assert len(move_group) == 3

    new_move_group = ot3utils.add_delay_to_move_group(
        move_group, present_nodes, ([NodeId.gantry_x], 3.0)
    )
    # new group should be one step longer
    assert len(new_move_group) == 4
    # all nodes have the same duration of all steps
    for node in present_nodes:
        assert sum([step[node].duration_sec for step in new_move_group]) == 6.0

    assert new_move_group[0][NodeId.gantry_x].duration_sec == 3.0
    assert new_move_group[0][NodeId.gantry_y].duration_sec == 1.0
    assert new_move_group[0][NodeId.head_l].duration_sec == 1.0
    assert new_move_group[3][NodeId.gantry_x].duration_sec == 1.0
    assert new_move_group[3][NodeId.gantry_y].duration_sec == 3.0
    assert new_move_group[3][NodeId.head_l].duration_sec == 3.0


def test_filter_zero_duration_step() -> None:
    origin = {
        Axis.X: 0,
        Axis.Y: 0,
        Axis.Z_L: 0,
        Axis.Z_R: 0,
        Axis.P_L: 0,
        Axis.P_R: 0,
    }
    moves = [Move.build_dummy([Axis.X, Axis.Y, Axis.Z_L, Axis.Z_R, Axis.P_L])]
    for block in (moves[0].blocks[0], moves[0].blocks[1]):
        block.distance = f64(25.0)
        block.time = f64(1.0)
        block.initial_speed = f64(25.0)
        block.acceleration = f64(0.0)
        block.final_speed = f64(25.0)
    present_nodes = [NodeId.gantry_x, NodeId.gantry_y, NodeId.head_l]
    move_group, final_pos = ot3utils.create_move_group(
        origin=origin,
        moves=moves,
        present_nodes=present_nodes,
    )
    assert len(move_group) == 2
    for step in move_group:
        assert set(present_nodes) == set(step.keys())


def test_nodeid_replace_head() -> None:
    assert ot3utils.replace_head_node(set([NodeId.head, NodeId.gantry_x])) == set(
        [NodeId.head_l, NodeId.head_r, NodeId.gantry_x]
    )
    assert ot3utils.replace_head_node(set([NodeId.gantry_x])) == set([NodeId.gantry_x])
    assert ot3utils.replace_head_node(set([NodeId.head_l])) == set([NodeId.head_l])


def test_nodeid_replace_gripper() -> None:
    assert ot3utils.replace_gripper_node(set([NodeId.gripper, NodeId.head])) == set(
        [NodeId.gripper_g, NodeId.gripper_z, NodeId.head]
    )
    assert ot3utils.replace_gripper_node(set([NodeId.head])) == set([NodeId.head])
    assert ot3utils.replace_gripper_node(set([NodeId.gripper_g])) == set(
        [NodeId.gripper_g]
    )


def test_get_system_contraints_for_plunger() -> None:
    set_acceleration = 2
    axis = Axis.P_L
    config = defaults_ot3.build_with_defaults({})
    updated_contraints = ot3utils.get_system_constraints_for_plunger_acceleration(
        config.motion_settings,
        conf_types.GantryLoad.LOW_THROUGHPUT,
        OT3Mount.LEFT,
        set_acceleration,
    )

    assert updated_contraints[axis].max_acceleration == set_acceleration


@pytest.mark.parametrize(["mount"], [[OT3Mount.LEFT], [OT3Mount.RIGHT]])
def test_get_system_constraints_for_emulsifying_pipette(mount: OT3Mount) -> None:
    set_max_speed = 90
    config = defaults_ot3.build_with_defaults({})
    pipette_ax = Axis.of_main_tool_actuator(mount)
    default_pip_max_speed = config.motion_settings.default_max_speed[
        conf_types.GantryLoad.LOW_THROUGHPUT
    ][OT3AxisKind.P]
    updated_constraints = ot3utils.get_system_constraints_for_emulsifying_pipette(
        config.motion_settings, conf_types.GantryLoad.LOW_THROUGHPUT, mount
    )
    other_pipette = list(set(Axis.pipette_axes()) - {pipette_ax})[0]
    assert updated_constraints[pipette_ax].max_speed == set_max_speed
    assert updated_constraints[other_pipette].max_speed == default_pip_max_speed


@pytest.mark.parametrize(
    ["moving", "expected"],
    [
        [
            [NodeId.gantry_x, NodeId.gantry_y, NodeId.gripper_g, NodeId.gripper_z],
            [],
        ],
        [
            [NodeId.head_l],
            [NodeId.pipette_left],
        ],
        [
            [NodeId.head_r],
            [NodeId.pipette_right],
        ],
    ],
)
def test_moving_pipettes_in_move_group(
    moving: List[NodeId], expected: List[NodeId]
) -> None:
    """Test that we can filter out the nonmoving nodes."""
    present_nodes = [
        NodeId.gantry_x,
        NodeId.gantry_y,
        NodeId.head_l,
        NodeId.head_r,
        NodeId.pipette_left,
        NodeId.pipette_right,
        NodeId.gripper_g,
        NodeId.gripper_z,
    ]

    moving_pipettes = ot3utils.moving_pipettes_in_move_group(
        set(present_nodes), set(moving)
    )
    assert set(moving_pipettes) == set(expected)



"""Run control command side-effect logic."""

import asyncio

from ..actions import ActionDispatcher, PauseAction, PauseSource
from ..state.state import StateStore


class RunControlHandler:
    """Implementation logic for protocol run control."""

    _state_store: StateStore

    def __init__(
        self,
        state_store: StateStore,
        action_dispatcher: ActionDispatcher,
    ) -> None:
        """Initialize a RunControlHandler instance."""
        self._state_store = state_store
        self._action_dispatcher = action_dispatcher

    async def wait_for_resume(self) -> None:
        """Issue a PauseAction to the store, pausing the run."""
        if not self._state_store.config.ignore_pause:
            self._action_dispatcher.dispatch(PauseAction(source=PauseSource.PROTOCOL))
            await self._state_store.wait_for(
                condition=self._state_store.commands.get_is_running
            )

    async def wait_for_duration(self, seconds: float) -> None:
        """Delay protocol execution for a duration."""
        if not self._state_store.config.ignore_pause:
            await asyncio.sleep(seconds)

    async def wait_for_tasks(self, tasks: list[str]) -> None:
        """Wait for concurrent tasks to complete."""
        await self._state_store.wait_for(
            condition=lambda: self._state_store.tasks.all_tasks_finished_or_any_task_failed(
                task_ids=tasks
            )
        )



import pytest

from opentrons.legacy_commands import protocol_commands


@pytest.mark.parametrize(
    argnames="seconds,minutes,expected_seconds,expected_minutes,expected_text",
    argvalues=[
        [10, 0, 10, 0, "Delaying for 0 minutes and 10.0 seconds"],
        [10, 9, 10, 9, "Delaying for 9 minutes and 10.0 seconds"],
        [100, 0, 40, 1, "Delaying for 1 minutes and 40.0 seconds"],
        [105, 5.25, 0, 7, "Delaying for 7 minutes and 0.0 seconds"],
        [0.5, 0, 0.5, 0, "Delaying for 0 minutes and 0.5 seconds"],
        [105.5, 5.25, 0.5, 7, "Delaying for 7 minutes and 0.5 seconds"],
        [0.998, 0, 0.998, 0, "Delaying for 0 minutes and 0.998 seconds"],
        [0.9998, 0, 0.9998, 0, "Delaying for 0 minutes and 1.0 seconds"],
        [1.0001, 0, 1.0001, 0, "Delaying for 0 minutes and 1.0 seconds"],
    ],
)
def test_delay(
    seconds: int,
    minutes: int,
    expected_seconds: int,
    expected_minutes: int,
    expected_text: str,
) -> None:
    command = protocol_commands.delay(seconds, minutes)
    name = command["name"]
    payload = command["payload"]

    assert name == "command.DELAY"
    assert payload["seconds"] == expected_seconds
    assert payload["minutes"] == expected_minutes
    assert payload["text"] == expected_text


def test_delay_with_message() -> None:
    """It should allow a message to be appended to the delay text."""
    command = protocol_commands.delay(seconds=1, minutes=1, msg="Waiting...")

    assert command["payload"]["text"] == (
        "Delaying for 1 minutes and 1.0 seconds. Waiting..."
    )



"""Tests for LiquidClass methods."""

import pytest
from decoy import Decoy

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
)

from opentrons.protocol_api import InstrumentContext, Labware, LiquidClass
from opentrons.protocol_api._liquid_properties import build_transfer_properties
from opentrons.protocols.advanced_control.transfers.common import (
    NoLiquidClassPropertyError,
)


def test_create_liquid_class(
    minimal_liquid_class_def1: LiquidClassSchemaV1,
) -> None:
    """It should create a LiquidClass from the provided definition."""
    assert LiquidClass.create(minimal_liquid_class_def1) == LiquidClass(
        _name="water1", _display_name="water 1", _by_pipette_setting={}
    )


def test_get_for_pipette_and_tip(
    decoy: Decoy,
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> None:
    """It should get the properties for the specified pipette and tip."""
    liq_class = LiquidClass.create(minimal_liquid_class_def2)
    result = liq_class.get_for("flex_1channel_50", "opentrons_flex_96_tiprack_50ul")
    assert result.aspirate.flow_rate_by_volume.as_dict() == {
        10.0: 40.0,
        20.0: 30.0,
    }
    mock_instrument = decoy.mock(cls=InstrumentContext)
    mock_tiprack = decoy.mock(cls=Labware)
    decoy.when(mock_instrument.name).then_return("flex_1channel_50")
    decoy.when(mock_tiprack.uri).then_return("opentrons_flex_96_tiprack_50ul")
    result_2 = liq_class.get_for(mock_instrument, mock_tiprack)
    assert result_2.aspirate.flow_rate_by_volume.as_dict() == {
        10.0: 40.0,
        20.0: 30.0,
    }


def test_get_for_raises_for_incorrect_pipette_or_tip(
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> None:
    """It should raise an error when accessing non-existent properties."""
    liq_class = LiquidClass.create(minimal_liquid_class_def2)

    with pytest.raises(NoLiquidClassPropertyError):
        liq_class.get_for("flex_1channel_50", "no_such_tiprack")

    with pytest.raises(NoLiquidClassPropertyError):
        liq_class.get_for("no_such_pipette", "opentrons_flex_96_tiprack_50ul")


def test_create_liquid_class_from_transfer_props(
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> None:
    """Should create a new liquid class from given data."""
    transfer_props = build_transfer_properties(
        minimal_liquid_class_def2.byPipette[0].byTipType[0]
    )
    by_pip_setting = {"my_pipette": {"my_tiprack": transfer_props}}
    new_liquid_class = LiquidClass.create_from(
        name="foo",
        display_name="bar",
        by_pipette_setting=by_pip_setting,
    )
    assert new_liquid_class.name == "foo"
    assert new_liquid_class.display_name == "bar"
    assert new_liquid_class.get_for("my_pipette", "my_tiprack") == transfer_props

    with pytest.raises(NoLiquidClassPropertyError):
        new_liquid_class.get_for("my_pipette", "no_such_tiprack")


def test_update_for_liquid_class_props(
    minimal_liquid_class_def2: LiquidClassSchemaV1,
    maximal_liquid_class_def: LiquidClassSchemaV1,
) -> None:
    """Should update the properties of the mentioned pipette and tip rack."""
    liq_class = LiquidClass.create(minimal_liquid_class_def2)
    assert (
        liq_class.get_for(
            "flex_1channel_50", "opentrons_flex_96_tiprack_50ul"
        ).aspirate.mix.enabled
        is False
    )

    sample_transfer_props = build_transfer_properties(
        maximal_liquid_class_def.byPipette[0].byTipType[0]
    )
    liq_class.update_for(
        pipette="flex_1channel_50",
        tip_rack="opentrons_flex_96_tiprack_50ul",
        transfer_properties=sample_transfer_props,
    )
    assert (
        liq_class.get_for("flex_1channel_50", "opentrons_flex_96_tiprack_50ul")
        == sample_transfer_props
    )
    assert (
        liq_class.get_for(
            "flex_1channel_50", "opentrons_flex_96_tiprack_50ul"
        ).aspirate.mix.enabled
        is True
    )



"""Test Flex Stacker empty command implementation."""

from typing import cast

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.flex_stacker.empty import (
    EmptyImpl,
    EmptyParams,
    EmptyResult,
)
from opentrons.protocol_engine.errors import (
    FlexStackerLabwarePoolNotYetDefinedError,
    ModuleNotLoadedError,
)
from opentrons.protocol_engine.execution import EquipmentHandler, RunControlHandler
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    BatchLabwareLocationUpdate,
    FlexStackerStateUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    DeckSlotLocation,
    InStackerHopperLocation,
    LabwareUri,
    NotOnDeckLocationSequenceComponent,
    StackerFillEmptyStrategy,
    StackerStoredLabwareGroup,
)
from opentrons.types import DeckSlotName


def _contained_labware(count: int) -> list[StackerStoredLabwareGroup]:
    return [
        StackerStoredLabwareGroup(
            primaryLabwareId=f"primary-id-{i + 1}",
            adapterLabwareId=None,
            lidLabwareId=None,
        )
        for i in range(count)
    ]


@pytest.fixture
def subject(
    state_view: StateView, run_control: RunControlHandler, equipment: EquipmentHandler
) -> EmptyImpl:
    """An EmptyImpl for testing."""
    return EmptyImpl(
        state_view=state_view, run_control=run_control, equipment=equipment
    )


@pytest.mark.parametrize(
    "current_stored,count_param,target_stored,removed",
    [
        pytest.param([], 0, [], [], id="empty-to-empty"),
        pytest.param(
            _contained_labware(3), 0, [], _contained_labware(3), id="full-to-empty"
        ),
        pytest.param(
            _contained_labware(3), 3, _contained_labware(3), [], id="full-noop"
        ),
        pytest.param(
            _contained_labware(2),
            3,
            _contained_labware(2),
            [],
            id="cant-increase",
        ),
        pytest.param(
            _contained_labware(3),
            2,
            _contained_labware(2),
            [_contained_labware(3)[-1]],
            id="not-full-empty",
        ),
        pytest.param(
            _contained_labware(3), 4, _contained_labware(3), [], id="overfull"
        ),
        pytest.param(
            _contained_labware(3), None, [], _contained_labware(3), id="default-count"
        ),
    ],
)
async def test_empty_happypath(
    decoy: Decoy,
    state_view: StateView,
    subject: EmptyImpl,
    current_stored: list[StackerStoredLabwareGroup],
    count_param: int | None,
    target_stored: list[StackerStoredLabwareGroup],
    removed: list[StackerStoredLabwareGroup],
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should empty a valid stacker's labware pool."""
    module_id = "some-module-id"
    stacker_state = FlexStackerSubState(
        module_id=cast(FlexStackerId, module_id),
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=current_stored,
        max_pool_count=3,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        stacker_state
    )
    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    params = EmptyParams(
        moduleId=module_id,
        count=count_param,
        message="some-message",
        strategy=StackerFillEmptyStrategy.LOGICAL,
    )
    result = await subject.execute(params)
    assert result.state_update == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id=module_id, contained_labware_bottom_first=target_stored
        ),
        batch_labware_location=BatchLabwareLocationUpdate(
            new_locations_by_id={
                g.primaryLabwareId: OFF_DECK_LOCATION for g in removed
            },
            new_offset_ids_by_id={g.primaryLabwareId: None for g in removed},
        ),
    )
    assert result.public == EmptyResult(
        count=len(target_stored),
        primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
        adapterLabwareURI=None,
        lidLabwareURI=None,
        storedLabware=target_stored,
        removedLabware=removed,
        originalPrimaryLabwareLocationSequences=[
            [InStackerHopperLocation(moduleId="some-module-id")] for _ in removed
        ],
        originalAdapterLabwareLocationSequences=None,
        originalLidLabwareLocationSequences=None,
        newPrimaryLabwareLocationSequences=[
            [NotOnDeckLocationSequenceComponent(logicalLocationName=OFF_DECK_LOCATION)]
            for _ in removed
        ],
        newAdapterLabwareLocationSequences=None,
        newLidLabwareLocationSequences=None,
    )


async def test_empty_requires_stacker(
    decoy: Decoy, state_view: StateView, subject: EmptyImpl
) -> None:
    """It should require a stacker."""
    decoy.when(state_view.modules.get_flex_stacker_substate("asda")).then_raise(
        ModuleNotLoadedError(module_id="asda")
    )
    with pytest.raises(ModuleNotLoadedError):
        await subject.execute(
            EmptyParams(
                moduleId="asda",
                strategy=StackerFillEmptyStrategy.LOGICAL,
                message="blah",
                count=3,
            )
        )


async def test_empty_requires_constrained_pool(
    decoy: Decoy, state_view: StateView, subject: EmptyImpl
) -> None:
    """It should require a constrained labware pool."""
    module_id = "module-id"
    stacker_state = FlexStackerSubState(
        module_id=cast(FlexStackerId, module_id),
        pool_primary_definition=None,
        pool_lid_definition=None,
        pool_adapter_definition=None,
        contained_labware_bottom_first=_contained_labware(3),
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        stacker_state
    )
    decoy.when(state_view.modules.get_location(module_id)).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_A3)
    )
    with pytest.raises(
        FlexStackerLabwarePoolNotYetDefinedError,
        match=".*The Flex Stacker in.*A3.*has not been configured yet and cannot be emptied.",
    ):
        await subject.execute(
            EmptyParams(
                moduleId=module_id,
                count=2,
                message="hello",
                strategy=StackerFillEmptyStrategy.LOGICAL,
            )
        )


async def test_pause_strategy_pauses(
    decoy: Decoy,
    state_view: StateView,
    run_control: RunControlHandler,
    subject: EmptyImpl,
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should pause the system when the pause strategy is used."""
    module_id = "some-module-id"
    current_count = 2
    count_param = 1
    target_count = 1
    stacker_state = FlexStackerSubState(
        module_id=cast(FlexStackerId, module_id),
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(current_count),
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(state_view.modules.get_flex_stacker_substate(module_id)).then_return(
        stacker_state
    )
    decoy.when(
        state_view.labware.get_uri_from_definition(flex_50uL_tiprack)
    ).then_return(LabwareUri("opentrons/opentrons_flex_96_filtertiprack_50ul/1"))
    params = EmptyParams(
        moduleId=module_id,
        count=count_param,
        message="some-message",
        strategy=StackerFillEmptyStrategy.MANUAL_WITH_PAUSE,
    )
    primary_id = _contained_labware(2)[1].primaryLabwareId
    result = await subject.execute(params)
    assert result.state_update == StateUpdate(
        flex_stacker_state_update=FlexStackerStateUpdate(
            module_id=module_id,
            contained_labware_bottom_first=_contained_labware(count_param),
        ),
        batch_labware_location=BatchLabwareLocationUpdate(
            new_locations_by_id={primary_id: OFF_DECK_LOCATION},
            new_offset_ids_by_id={primary_id: None},
        ),
    )
    assert result.public == EmptyResult(
        count=target_count,
        primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
        adapterLabwareURI=None,
        lidLabwareURI=None,
        storedLabware=_contained_labware(1),
        removedLabware=_contained_labware(2)[1:],
        originalPrimaryLabwareLocationSequences=[
            [InStackerHopperLocation(moduleId="some-module-id")]
        ],
        originalAdapterLabwareLocationSequences=None,
        originalLidLabwareLocationSequences=None,
        newPrimaryLabwareLocationSequences=[
            [NotOnDeckLocationSequenceComponent(logicalLocationName=OFF_DECK_LOCATION)]
        ],
        newAdapterLabwareLocationSequences=None,
        newLidLabwareLocationSequences=None,
    )
    decoy.verify(await run_control.wait_for_resume())



"""Command models to close the latch of a Flex Stacker."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional

from pydantic import BaseModel, Field
from typing_extensions import Type

from ...errors import ErrorOccurrence
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData

if TYPE_CHECKING:
    from ...state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

UnsafeFlexStackerCloseLatchCommandType = Literal["unsafe/flexStacker/closeLatch"]


class UnsafeFlexStackerCloseLatchParams(BaseModel):
    """The parameters defining how a stacker should close its latch."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")


class UnsafeFlexStackerCloseLatchResult(BaseModel):
    """Result data from a stacker UnsafeFlexStackerCloseLatch command."""


class UnsafeFlexStackerCloseLatchImpl(
    AbstractCommandImpl[
        UnsafeFlexStackerCloseLatchParams,
        SuccessData[UnsafeFlexStackerCloseLatchResult],
    ]
):
    """Implementation of a stacker UnsafeFlexStackerCloseLatch command."""

    def __init__(
        self, state_view: StateView, equipment: EquipmentHandler, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: UnsafeFlexStackerCloseLatchParams
    ) -> SuccessData[UnsafeFlexStackerCloseLatchResult]:
        """Execute the stacker UnsafeFlexStackerCloseLatch command.

        Closing the latch modifies the state of the flex stacker and affects its
        ability to execute the next command. This command should be used with care
        outside of a protocol.
        """
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        if stacker_hw is not None:
            await stacker_hw.close_latch()
        return SuccessData(public=UnsafeFlexStackerCloseLatchResult())


class UnsafeFlexStackerCloseLatch(
    BaseCommand[
        UnsafeFlexStackerCloseLatchParams,
        UnsafeFlexStackerCloseLatchResult,
        ErrorOccurrence,
    ]
):
    """A command to UnsafeFlexStackerCloseLatch the Flex Stacker of labware."""

    commandType: UnsafeFlexStackerCloseLatchCommandType = (
        "unsafe/flexStacker/closeLatch"
    )
    params: UnsafeFlexStackerCloseLatchParams
    result: Optional[UnsafeFlexStackerCloseLatchResult] = None

    _ImplementationCls: Type[UnsafeFlexStackerCloseLatchImpl] = (
        UnsafeFlexStackerCloseLatchImpl
    )


class UnsafeFlexStackerCloseLatchCreate(
    BaseCommandCreate[UnsafeFlexStackerCloseLatchParams]
):
    """A request to execute a Flex Stacker UnsafeFlexStackerCloseLatch command."""

    commandType: UnsafeFlexStackerCloseLatchCommandType = (
        "unsafe/flexStacker/closeLatch"
    )
    params: UnsafeFlexStackerCloseLatchParams

    _CommandCls: Type[UnsafeFlexStackerCloseLatch] = UnsafeFlexStackerCloseLatch
